# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

"""Test single point logfiles in cclib."""

import datetime
import os
import unittest

import numpy
import packaging

from common import get_minimum_carbon_separation

from skip import skipForParser
from skip import skipForLogfile


__filedir__ = os.path.realpath(os.path.dirname(__file__))


class GenericSPTest(unittest.TestCase):
    """Generic restricted single point unittest"""

    # Molecular mass of DVB in mD, and expected precision.
    molecularmass = 130078.25
    mass_precision = 0.10

    # In STO-3G, H has 1, C has 5 (1 S and 4 SP).
    nbasisdict = {1:1, 6:5}

    # Approximate B3LYP energy of dvb after SCF in STO-3G.
    b3lyp_energy = -10365

    # Approximate energy of the innermost molecular orbital of DVB with
    # B3LYP/STO-3G (from Q-Chem 5.4).
    b3lyp_moenergy = -272.60365543
    b3lyp_moenergy_delta = 7.55e-2

    # Overlap first two atomic orbitals.
    overlap01 = 0.24

    # Generally, one criteria for SCF energy convergence.
    num_scf_criteria = 1

    def testnatom(self):
        """Is the number of atoms equal to 20?"""
        assert self.data.natom == 20

    def testatomnos(self):
        """Are the atomnos correct?"""

        # The nuclear charges should be integer values in a NumPy array.
        assert numpy.alltrue([numpy.issubdtype(atomno, numpy.signedinteger)
                                       for atomno in self.data.atomnos])
        assert self.data.atomnos.dtype.char == 'i'

        assert self.data.atomnos.shape == (20,)
        assert sum(self.data.atomnos == 6) + sum(self.data.atomnos == 1) == 20

    @skipForParser('DALTON', 'DALTON has a very low accuracy for the printed values of all populations (2 decimals rounded in a weird way), so let it slide for now')
    @skipForParser('FChk', 'The parser is still being developed so we skip this test')
    @skipForParser('GAMESSDAT', 'We are not going to implement atom charges in this version.')
    @skipForLogfile('Jaguar/basicJaguar7', 'We did not print the atomic partial charges in the unit tests for this version')
    @skipForLogfile('Molpro/basicMolpro2006', "These tests were run a long time ago and since we don't have access to Molpro 2006 anymore, we can skip this test (it is tested in 2012)")
    @skipForParser('Turbomole','The parser is still being developed so we skip this test')
    def testatomcharges(self):
        """Are atomic charges consistent with natom?"""
        for atomcharge_type in self.data.atomcharges:
            charges = self.data.atomcharges[atomcharge_type]
            natom = self.data.natom
            assert len(charges) == natom, f"len(atomcharges['{atomcharge_type}']) = {len(charges)}, natom = {natom}"

    @skipForParser('DALTON', 'DALTON has a very low accuracy for the printed values of all populations (2 decimals rounded in a weird way), so let it slide for now')
    @skipForLogfile("FChk/basicQChem5.2", "not printed for Q-Chem")
    @skipForLogfile("FChk/basicQChem5.4", "not printed for Q-Chem")
    @skipForParser('GAMESSDAT', 'We are not sure about the specific type of atom charges, it is best to skip the test for now.')
    @skipForLogfile('Jaguar/basicJaguar7', 'We did not print the atomic partial charges in the unit tests for this version')
    @skipForLogfile('Molpro/basicMolpro2006', "These tests were run a long time ago and since we don't have access to Molpro 2006 anymore, we can skip this test (it is tested in 2012)")
    @skipForParser('Turbomole','The parser is still being developed so we skip this test')
    def testatomcharges_mulliken(self):
        """Do Mulliken atomic charges sum to zero?"""
        charges = self.data.atomcharges["mulliken"]
        assert abs(sum(charges)) < 1.0e-2

    @skipForParser('ADF', 'Lowdin charges not present by default')
    @skipForParser('DALTON', 'DALTON has a very low accuracy for the printed values of all populations (2 decimals rounded in a weird way), so let it slide for now')
    @skipForParser('FChk', 'The parser is still being developed so we skip this test')
    @skipForParser('GAMESSDAT', 'We are not sure about the specific type of atom charges, it is best to skip the test for now.')
    @skipForParser('Gaussian', 'Lowdin charges not present by default')
    @skipForParser('Jaguar', 'Lowdin charges not present by default')
    @skipForParser('NWChem', 'Lowdin charges not present by default')
    @skipForParser('Molcas', 'Lowdin charges not present by default')
    @skipForParser('Molpro', 'Lowdin charges not present by default')
    @skipForParser('QChem', 'Lowdin charges not present by default')
    @skipForParser('Turbomole','The parser is still being developed so we skip this test')
    def testatomcharges_lowdin(self):
        """Do Lowdin atomic charges sum to zero?"""
        charges = self.data.atomcharges["lowdin"]
        assert abs(sum(charges)) < 1.0e-2

    @skipForParser('ADF', 'Hirshfeld charges not implemented')
    @skipForParser('DALTON', 'DALTON has a very low accuracy for the printed values of all populations (2 decimals rounded in a weird way), so let it slide for now')
    @skipForParser('FChk', 'The parser is still being developed so we skip this test')
    @skipForParser('Gaussian', 'Hirshfeld charges not implemented')
    @skipForParser('GAMESS', 'Hirshfeld charges not implemented')
    @skipForParser('GAMESSUK', 'Hirshfeld charges not implemented')
    @skipForParser('GAMESSDAT', 'We are not sure about the specific type of atom charges, it is best to skip the test for now.')
    @skipForParser('Jaguar', 'Hirshfeld charges not implemented')
    @skipForParser('NWChem', 'Hirshfeld charges not implemented')
    @skipForParser('Molcas', 'Hirshfeld charges not implemented')
    @skipForParser('Molpro', 'Hirshfeld charges not implemented')
    @skipForLogfile('ORCA/basicORCA4.1', 'This needs to be moved to regressions')
    @skipForParser('Psi4', 'Hirshfeld charges not implemented')
    @skipForParser('QChem', 'Hirshfeld charges not implemented')
    @skipForParser('Turbomole','The parser is still being developed so we skip this test')
    def testatomcharges_hirshfeld(self):
        """Do Hirshfeld atomic charges sum to roughly zero?"""
        charges = self.data.atomcharges["hirshfeld"]
        assert abs(sum(charges)) < 4.0e-3

    def testatomcoords(self):
        """Are the dimensions of atomcoords 1 x natom x 3?"""
        expected_shape = (1, self.data.natom, 3)
        assert self.data.atomcoords.shape == expected_shape

    @skipForParser('GAMESSDAT', 'Vectors need some calculations to transform them. Current mm value is 2.54')
    def testatomcoords_units(self):
        """Are atomcoords consistent with Angstroms?"""
        min_carbon_dist = get_minimum_carbon_separation(self.data)
        dev = abs(min_carbon_dist - 1.34)
        assert dev < 0.03, f"Minimum carbon dist is {min_carbon_dist:.2f} (not 1.34)"

    @skipForParser('GAMESSDAT', 'Neither charge nor mult exists in the files.')
    @skipForParser('Molcas', 'missing mult')
    def testcharge_and_mult(self):
        """Are the charge and multiplicity correct?"""
        assert self.data.charge == 0
        assert self.data.mult == 1

    def testnbasis(self):
        """Is the number of basis set functions correct?"""
        count = sum([self.nbasisdict[n] for n in self.data.atomnos])
        assert self.data.nbasis == count

    @skipForParser('ADF', 'ADF parser does not extract atombasis')
    @skipForLogfile('Jaguar/basicJaguar7', 'Data file does not contain enough information. Can we make a new one?')
    @skipForParser('Molcas','The parser is still being developed so we skip this test')
    @skipForParser('Turbomole','The parser is still being developed so we skip this test')
    def testatombasis(self):
        """Are the indices in atombasis the right amount and unique?"""
        all = []
        for i, atom in enumerate(self.data.atombasis):
            assert len(atom) == self.nbasisdict[self.data.atomnos[i]]
            all += atom
        # Test if there are as many indices as atomic orbitals.
        assert len(all) == self.data.nbasis
        # Check if all are different (every orbital indexed once).
        assert len(set(all)) == len(all)

    @skipForLogfile("FChk/basicQChem5.2", "Q-Chem doesn't print SCF energy to fchk")
    @skipForLogfile("FChk/basicQChem5.4", "Q-Chem doesn't print SCF energy to fchk")
    @skipForParser('GAMESS', 'atommasses not implemented yet')
    @skipForParser('GAMESSUK', 'atommasses not implemented yet')
    @skipForParser('GAMESSDAT', 'Atommasses implemented, but it does not pass the test.')
    @skipForParser('Jaguar', 'atommasses not implemented yet')
    @skipForParser('Molcas','The parser is still being developed so we skip this test')
    @skipForParser('Molpro', 'atommasses not implemented yet')
    @skipForLogfile('Psi4/basicPsi4.0b5', 'atommasses not implemented yet')
    @skipForParser('QChem', 'atommasses not implemented yet')
    @skipForParser('Turbomole','The parser is still being developed so we skip this test')
    def testatommasses(self):
        """Do the atom masses sum up to the molecular mass?"""
        mm = 1000*sum(self.data.atommasses)
        msg = f"Molecule mass: {mm:f} not {self.molecularmass:f} +- {self.mass_precision:f}mD"
        assert abs(mm-self.molecularmass) < self.mass_precision, msg

    @skipForParser('Turbomole','The parser is still being developed so we skip this test')
    def testcoreelectrons(self):
        """Are the coreelectrons all 0?"""
        ans = numpy.zeros(self.data.natom, 'i')
        numpy.testing.assert_array_equal(self.data.coreelectrons, ans)

    @skipForParser('FChk', 'Formatted checkpoint files do not have a section for symmetry')
    @skipForParser('GAMESSDAT', 'Mosyms do not exist in the file')
    @skipForParser('Molcas','The parser is still being developed so we skip this test')
    @skipForParser('Molpro', '?')
    def testsymlabels(self):
        """Are all the symmetry labels either Ag/u or Bg/u?"""
        sumwronglabels = sum([x not in ['Ag', 'Bu', 'Au', 'Bg'] for x in self.data.mosyms[0]])
        assert sumwronglabels == 0

    def testhomos(self):
        """Is the index of the HOMO equal to 34?"""
        numpy.testing.assert_array_equal(
            self.data.homos,
            numpy.array([34], "i"),
            f"{numpy.array_repr(self.data.homos)} != array([34],'i')",
        )

    @skipForParser('FChk', 'Formatted Checkpoint files do not have a section for SCF energy')
    @skipForParser('GAMESSDAT', 'Scfvalues probably do not exist in the file')
    def testscfvaluetype(self):
        """Are scfvalues and its elements the right type??"""
        assert isinstance(self.data.scfvalues, list)
        assert isinstance(self.data.scfvalues[0], numpy.ndarray)

    @skipForLogfile("FChk/basicQChem5.2", "Q-Chem doesn't print SCF energy to fchk")
    @skipForLogfile("FChk/basicQChem5.4", "Q-Chem doesn't print SCF energy to fchk")
    @skipForParser('GAMESSDAT', 'Scfenergies probably do not exist in the file')
    def testscfenergy(self):
        """Is the SCF energy within the target?"""
        assert abs(self.data.scfenergies[-1]-self.b3lyp_energy) < 40

    @skipForParser('FChk', 'Formatted Checkpoint files do not have a section for SCF convergence')
    @skipForParser('GAMESSDAT', 'Scftargets probably do not exist in the file')
    def testscftargetdim(self):
        """Do the scf targets have the right dimensions?"""
        assert self.data.scftargets.shape == (len(self.data.scfvalues), len(self.data.scfvalues[0][0]))

    @skipForParser('FChk', 'Formatted Checkpoint files do not have a section for SCF convergence')
    @skipForParser('GAMESSDAT', 'Scftargets probably do not exist in the file')
    def testscftargets(self):
        """Are correct number of SCF convergence criteria being parsed?"""
        assert len(self.data.scftargets[0]) == self.num_scf_criteria

    def testlengthmoenergies(self):
        """Is the number of evalues equal to nmo?"""
        if hasattr(self.data, "moenergies"):
            assert len(self.data.moenergies[0]) == self.data.nmo

    def testtypemoenergies(self):
        """Is moenergies a list containing one numpy array?"""
        if hasattr(self.data, "moenergies"):
            assert isinstance(self.data.moenergies, list)
            assert isinstance(self.data.moenergies[0], numpy.ndarray)

    @skipForParser('GAMESSDAT', 'Moenergies probably do not exist in the file')
    @skipForLogfile('Gaussian/basicGaussian16/dvb_sp_no.out', 'no energies for natural orbitals')
    @skipForLogfile('Turbomole/basicTurbomole5.9/dvb_sp_symm', 'delta of 7.4, everything else ok')
    def testfirstmoenergy(self):
        """Is the lowest energy molecular orbital within the target?"""
        assert abs(self.data.moenergies[0][0]-self.b3lyp_moenergy) < self.b3lyp_moenergy_delta

    @skipForParser('DALTON', 'mocoeffs not implemented yet')
    @skipForLogfile('Jaguar/basicJaguar7', 'Data file does not contain enough information. Can we make a new one?')
    @skipForParser('Turbomole', 'Use of symmetry has reduced the number of mo coeffs')
    def testdimmocoeffs(self):
        """Are the dimensions of mocoeffs equal to 1 x nmo x nbasis?"""
        if hasattr(self.data, "mocoeffs"):
            assert isinstance(self.data.mocoeffs, list)
            assert len(self.data.mocoeffs) == 1
            assert self.data.mocoeffs[0].shape == (self.data.nmo, self.data.nbasis)

    @skipForParser('DALTON', 'mocoeffs not implemented yet')
    @skipForLogfile('Jaguar/basicJaguar7', 'Data file does not contain enough information. Can we make a new one?')
    def testfornoormo(self):
        """Do we have NOs or MOs?"""
        assert hasattr(self.data, "nocoeffs") or hasattr(self.data, "mocoeffs")

    def testdimnoccnos(self):
        """Is the length of nooccnos equal to nmo?"""
        if hasattr(self.data, "nooccnos"):
            assert isinstance(self.data.nooccnos, numpy.ndarray)
            assert len(self.data.nooccnos) == self.data.nmo

    def testdimnocoeffs(self):
        """Are the dimensions of nocoeffs equal to nmo x nmo?"""
        if hasattr(self.data, "nocoeffs"):
            assert isinstance(self.data.nocoeffs, numpy.ndarray)
            assert self.data.nocoeffs.shape == (self.data.nmo, self.data.nmo)

    @skipForParser('DALTON', 'To print: **INTEGRALS\n.PROPRI')
    @skipForLogfile('FChk/basicGaussian09', 'Only available in QChem')
    @skipForLogfile('FChk/basicGaussian16', 'Only available in QChem')
    @skipForParser('GAMESSDAT','Aooverlaps probably do not exist in the file.')
    @skipForParser('Molcas','The parser is still being developed so we skip this test')
    @skipForParser('Psi4', 'Psi4 does not currently have the option to print the overlap matrix')
    @skipForParser('QChem', 'QChem cannot print the overlap matrix')
    @skipForParser('Turbomole','The parser is still being developed so we skip this test')
    def testaooverlaps(self):
        """Are the dims and values of the overlap matrix correct?"""

        assert self.data.aooverlaps.shape == (self.data.nbasis, self.data.nbasis)

        # The matrix is symmetric.
        row = self.data.aooverlaps[0,:]
        col = self.data.aooverlaps[:,0]
        assert sum(col - row) == 0.0

        # All values on diagonal should be exactly one.
        for i in range(self.data.nbasis):
            assert self.data.aooverlaps[i,i] == 1.0

        # Check some additional values that don't seem to move around between programs.
        assert abs(self.data.aooverlaps[0, 1] - self.overlap01) < 0.01
        assert abs(self.data.aooverlaps[1, 0] - self.overlap01) < 0.01
        assert round(abs(self.data.aooverlaps[3, 0]), 7) == 0
        assert round(abs(self.data.aooverlaps[0, 3]), 7) == 0

    def testoptdone(self):
        """There should be no optdone attribute set."""
        assert not hasattr(self.data, 'optdone')

    @skipForParser('ADF', 'Not implemented yes')
    @skipForParser('DALTON', 'Not implemented yes')
    @skipForParser('FChk', 'Rotational constants are never written to fchk files')
    @skipForParser('GAMESS', 'Not implemented yes')
    @skipForParser('GAMESSUK', 'Not implemented yet')
    @skipForParser('GAMESSDAT', 'Not implemented yet')
    @skipForParser('Jaguar', 'Not implemented yet')
    @skipForParser('Molcas', 'Not implemented yes')
    @skipForParser('Molpro', 'Not implemented yes')
    @skipForParser('NWChem', 'Not implemented yes')
    @skipForParser('ORCA', 'Not implemented yes')
    @skipForParser('Psi4', 'Not implemented yes')
    @skipForParser('QChem', 'Not implemented yes')
    @skipForParser('Turbomole', 'Not implemented yes')
    def testrotconsts(self):
        """A single geometry leads to single set of rotational constants."""
        assert self.data.rotconsts.shape == (1, 3)
        # taken from Gaussian16/dvb_sp.out
        ref = [4.6266363, 0.6849065, 0.5965900]
        numpy.testing.assert_allclose(self.data.rotconsts[0], ref, rtol=0, atol=1.0e-3)

    @skipForParser('FChk', 'The parser is still being developed so we skip this test')
    @skipForParser('Gaussian', 'Logfile needs to be updated')
    @skipForParser('Jaguar', 'No dipole moments in the logfile')
    @skipForParser('Molcas','The parser is still being developed so we skip this test')
    def testmoments(self):
        """Does the dipole and possible higher molecular moments look reasonable?"""

        # The reference point is always a vector, but not necessarily the
        # origin or center of mass. In this case, however, the center of mass
        # is at the origin, so we now what to expect.
        reference = self.data.moments[0]
        assert len(reference) == 3
        for x in reference:
            assert x == 0.0

        # Length and value of dipole moment should always be correct (zero for this test).
        dipole = self.data.moments[1]
        assert len(dipole) == 3
        for d in dipole:
            assert round(abs(d), 7) == 0

        # If the quadrupole is there, we can expect roughly -50B for the XX moment,
        # -50B for the YY moment and and -60B for the ZZ moment.
        if len(self.data.moments) > 2:
            quadrupole = self.data.moments[2]
            assert len(quadrupole) == 6
            assert abs(quadrupole[0] - -50) < 2.5
            assert abs(quadrupole[3] - -50) < 2.5
            assert abs(quadrupole[5] - -60) < 3

        # If the octupole is there, it should have 10 components and be zero.
        if len(self.data.moments) > 3:
            octupole = self.data.moments[3]
            assert len(octupole) == 10
            for m in octupole:
                assert abs(m) < 0.001

        # The hexadecapole should have 15 elements, an XXXX component of around -1900 Debye*ang^2,
        # a YYYY component of -330B and a ZZZZ component of -50B.
        if len(self.data.moments) > 4:
            hexadecapole = self.data.moments[4]
            assert len(hexadecapole) == 15
            assert abs(hexadecapole[0] - -1900) < 90
            assert abs(hexadecapole[10] - -330) < 11
            assert abs(hexadecapole[14] - -50) < 2.5

        # The are 21 unique 32-pole moments, and all are zero in this test case.
        if len(self.data.moments) > 5:
            moment32 = self.data.moments[5]
            assert len(moment32) == 21
            for m in moment32:
                assert m == 0.0

    @skipForParser('ADF', 'reading basis set names is not implemented')
    @skipForParser('GAMESSDAT', 'Basis set not implemented in this version.')
    @skipForParser('GAMESSUK', 'reading basis set names is not implemented')
    @skipForParser('Molcas', 'reading basis set names is not implemented')
    @skipForParser('Psi4', 'reading basis set names is not implemented')
    def testmetadata_basis_set(self):
        """Does metadata have expected keys and values?"""
        assert self.data.metadata["basis_set"].lower() == "sto-3g"

    @skipForParser('ADF', 'reading input file contents and name is not implemented')
    @skipForParser('DALTON', 'reading input file contents and name is not implemented')
    @skipForParser('FChk', 'Formatted checkpoint files do not have an input file section')
    @skipForParser('GAMESS', 'reading input file contents and name is not implemented')
    @skipForParser('GAMESSUK', 'reading input file contents and name is not implemented')
    @skipForParser('GAMESSDAT', 'reading input file contents and name is not implemented')
    @skipForParser('Gaussian', 'reading input file contents and name is not implemented')
    @skipForParser('Jaguar', 'reading input file contents and name is not implemented')
    @skipForParser('Molcas', 'reading input file contents and name is not implemented')
    @skipForParser('Molpro', 'reading input file contents and name is not implemented')
    @skipForParser('NWChem', 'reading input file contents and name is not implemented')
    @skipForParser('Psi4', 'reading input file contents and name is not implemented')
    @skipForParser('QChem', 'reading input file contents and name is not implemented')
    @skipForParser('Turbomole', 'reading input file contents and name is not implemented')
    def testmetadata_input_file(self):
        """Does metadata have expected keys and values?"""
        assert "input_file_contents" in self.data.metadata
        # TODO make input file names consistent where possible, though some
        # programs do not allow arbitrary file extensions; for example, DALTON
        # must end in `dal`.
        assert "dvb_sp.in" in self.data.metadata["input_file_name"]

    def testmetadata_methods(self):
        """Does metadata have expected keys and values?"""
        # TODO implement and unify across parsers; current values are [],
        # ["HF"], ["RHF"], and ["DFT"]
        assert "methods" in self.data.metadata

    def testmetadata_package(self):
        """Does metadata have expected keys and values?"""
        # TODO How can the value be tested when the package name comes from
        # the parser and isn't stored on ccData?
        assert "package" in self.data.metadata

    @skipForParser('FChk', 'Formatted Checkpoint files do not have section for legacy package version')
    @skipForParser('GAMESSDAT', 'Files do not contain information about the legacy package version')
    def testmetadata_legacy_package_version(self):
        """Does metadata have expected keys and values?"""
        # TODO Test specific values for each unit test.
        assert "legacy_package_version" in self.data.metadata

    @skipForParser('FChk', 'Formatted Checkpoint files do not have section for package version')
    @skipForParser('GAMESSDAT', 'Files do not contain information about the package version')
    def testmetadata_package_version(self):
        """Does metadata have expected keys and values?"""
        # TODO Test specific values for each unit test.
        assert isinstance(packaging.version.parse(self.data.metadata["package_version"]), packaging.version.Version)

    @skipForParser('FChk', 'point group symmetry cannot be printed')
    @skipForParser('GAMESSDAT', 'Files probably do not contain information about symmetry_detected')
    @skipForParser('Molcas', 'reading point group symmetry and name is not implemented')
    @skipForParser('Molpro', 'reading point group symmetry and name is not implemented')
    @skipForParser('Turbomole', 'reading point group symmetry and name is not implemented')
    def testmetadata_symmetry_detected(self):
        """Does metadata have expected keys and values?"""
        assert self.data.metadata["symmetry_detected"] == "c2h"

    @skipForParser('FChk', 'point group symmetry cannot be printed')
    @skipForParser('GAMESSDAT', 'Files probably do not contain information about symmetry_used')
    @skipForParser('Molcas', 'reading point group symmetry and name is not implemented')
    @skipForParser('Molpro', 'reading point group symmetry and name is not implemented')
    @skipForParser('Turbomole', 'reading point group symmetry and name is not implemented')
    def testmetadata_symmetry_used(self):
        """Does metadata have expected keys and values?"""
        assert self.data.metadata["symmetry_used"] == "c2h"

    @skipForParser('ADF', 'reading cpu/wall time is not implemented for this parser')
    @skipForParser('DALTON', 'reading cpu/wall time is not implemented for this parser') 
    @skipForParser('FChk', 'reading cpu/wall time is not implemented for this parser') 
    @skipForParser('GAMESS', 'reading cpu/wall time is not implemented for this parser') 
    @skipForParser('GAMESSUK', 'reading cpu/wall time is not implemented for this parser') 
    @skipForParser('GAMESSUS', 'reading cpu/wall time is not implemented for this parser') 
    @skipForParser('GAMESSDAT', 'reading cpu/wall time is not implemented for this parser')
    @skipForParser('Jaguar', 'reading cpu/wall time is not implemented for this parser') 
    @skipForParser('Molcas', ' reading cpu/wall time is not implemented for this parser') 
    @skipForParser('Molpro', 'reading cpu/wall time is not implemented for this parser') 
    @skipForParser('NWChem', 'reading cpu/wall time is not implemented for this parser')  
    @skipForParser('Psi3', 'reading cpu/wall time is not implemented for this parser') 
    @skipForParser('Psi4', 'reading cpu/wall time is not implemented for this parser')
    def testmetadata_times(self):
        """Does metadata have expected keys and values of correct types?"""
        if "wall_time" in self.data.metadata:
            assert self.data.metadata["wall_time"]
            assert all(isinstance(wall_time, datetime.timedelta)
                       for wall_time in self.data.metadata["wall_time"])
        if "cpu_time" in self.data.metadata:
            assert self.data.metadata["cpu_time"]
            assert all(isinstance(cpu_time, datetime.timedelta)
                       for cpu_time in self.data.metadata["cpu_time"])


class GenericHFSPTest(GenericSPTest):

    # Approximate HF energy of dvb after SCF in STO-3G (from DALTON 2015).
    hf_scfenergy = -10334.03948035995

    # Approximate energy of the innermost molecular orbital of DVB with
    # HF/STO-3G (from Psi4 1.3.1).
    hf_moenergy = -300.43401785663235

    @skipForParser('FChk', 'Formatted Checkpoint files do not have a section for SCF energy')
    def testscfenergy(self):
        """Is the SCF energy within the target?"""
        assert abs(self.data.scfenergies[-1]-self.hf_scfenergy) < 6.5e-1

    def testfirstmoenergy(self):
        """Is the lowest energy molecular orbital within the target?"""
        assert abs(self.data.moenergies[0][0]-self.hf_moenergy) < 1.6e-1


class ADFSPTest(GenericSPTest):
    """Customized restricted single point unittest"""

    # ADF only prints up to 0.1mD per atom, so the precision here is worse than 0.1mD.
    mass_precision = 0.3

    foverlap00 = 1.00003
    foverlap11 = 1.02672
    foverlap22 = 1.03585
    num_scf_criteria = 2
    b3lyp_energy = -140
    b3lyp_moenergy = -269.6079423873336

    def testfoverlaps(self):
        """Are the dims and values of the fragment orbital overlap matrix correct?"""

        assert self.data.fooverlaps.shape == (self.data.nbasis, self.data.nbasis)

        # The matrix is symmetric.
        row = self.data.fooverlaps[0,:]
        col = self.data.fooverlaps[:,0]
        assert sum(col - row) == 0.0

        # Although the diagonal elements are close to zero, the SFOs
        # are generally not normalized, so test for a few specific values.
        assert abs(self.data.fooverlaps[0, 0]-self.foverlap00) < 0.0001
        assert abs(self.data.fooverlaps[1, 1]-self.foverlap11) < 0.0001
        assert abs(self.data.fooverlaps[2, 2]-self.foverlap22) < 0.0001


class GaussianSPTest(GenericSPTest):
    """Customized restricted single point unittest"""

    num_scf_criteria = 3

class JaguarSPTest(GenericSPTest):
    """Customized restricted single point KS unittest"""

    num_scf_criteria = 2


class JaguarHFSPTest(JaguarSPTest, GenericHFSPTest):
    """Customized restricted single point KS unittest"""


class Jaguar7SPTest(JaguarSPTest):
    """Customized restricted single point unittest"""

    # Jaguar prints only 10 virtual MOs by default. Can we re-run with full output?
    def testlengthmoenergies(self):
        """Is the number of evalues equal to the number of occ. MOs + 10?"""
        assert len(self.data.moenergies[0]) == self.data.homos[0]+11

class MolcasSPTest(GenericSPTest):
    """Customized restricted single point unittest"""

    num_scf_criteria = 4

class MolproSPTest(GenericHFSPTest):
    """Customized restricted single point HF unittest"""

    num_scf_criteria = 2

class NWChemKSSPTest(GenericSPTest):
    """Customized restricted single point unittest"""

    num_scf_criteria = 3

class PsiSPTest(GenericSPTest):
    """Customized restricted single point KS unittest"""

    num_scf_criteria = 2


class PsiHFSPTest(PsiSPTest, GenericHFSPTest):
    """Customized restricted single point HF unittest"""


class OrcaSPTest(GenericSPTest):
    """Customized restricted single point unittest"""

    # Orca has different weights for the masses
    molecularmass = 130190

    b3lyp_moenergy_delta = 1.2e-1

    num_scf_criteria = 3


class TurbomoleSPTest(GenericSPTest):
    """Customized restricted single point KS unittest"""

    num_scf_criteria = 2
    
    def testmetadata_basis_set(self):
        """Does metadata have expected keys and values?"""
        # One of our test cases used sto-3g hondo
        valid_basis = self.data.metadata["basis_set"].lower() in ("sto-3g", "sto-3g hondo")
        assert valid_basis


class TurbomoleHFSPTest(TurbomoleSPTest, GenericHFSPTest):
    """Customized restricted single point HF unittest"""


class GenericDispersionTest(unittest.TestCase):
    """Generic single-geometry dispersion correction unittest"""

    dispersionenergy = -0.4005496

    def testdispersionenergies(self):
        """Is the dispersion energy parsed correctly?"""
        assert len(self.data.dispersionenergies) == 1
        assert abs(self.data.dispersionenergies[0]-self.dispersionenergy) < 2.0e-7


class FireflyDispersionTest(GenericDispersionTest):
    """Customized single-geometry dispersion correction unittest"""
    dispersionenergy = -0.4299821
    

class SolventMetadataTest(unittest.TestCase):
    """Check we can parse implicit solvent data."""
    
    model = ""
    # Toluene
    static_dielectric_constant = 2.3741
    
    def test_solvent_model(self) -> None:
        """Check solvent model was parsed correctly"""
        assert self.data.metadata['solvent_model'] == self.model

    @skipForLogfile("basicQChem6.0/water_hf_solvent_smd_iefpcm.out", "the internally-used dielectric constant isn't printed, only solvent name")
    def test_solvent_dielectric(self) -> None:
        """Check solvent dielectric was parsed correctly"""
        assert abs(self.data.metadata['solvent_params']['epsilon'] - self.static_dielectric_constant) < 1.0e-4


class QChemSolventMetadataTest(SolventMetadataTest):

    static_dielectric_constant = 2.370
        

class IEFPCMMetadataTest(SolventMetadataTest):
    """Check we can parse implicit solvent data."""

    model = "IEFPCM"
    
class SCIPCMMetadataTest(SolventMetadataTest):
    """Check we can parse implicit solvent data."""

    model = "SCIPCM"
    
class IPCMMetadataTest(SolventMetadataTest):
    """Check we can parse implicit solvent data."""

    model = "IPCM"
    static_dielectric_constant = 78.3
    
class COSMOMetadataTest(SolventMetadataTest):
    """Check we can parse implicit solvent data."""

    model = "COSMO"

class CPCMMetadataTest(SolventMetadataTest):
    """Check we can parse implicit solvent data."""

    model = "CPCM"
    
class CPCMCOSMOMetadataTest(SolventMetadataTest):
    """Check we can parse implicit solvent data."""

    model = "CPCM-COSMO"
    
class SMDIEFPCMMetadataTest(SolventMetadataTest):
    """Check we can parse implicit solvent data."""

    model = "SMD-IEFPCM"
    
class SMDCPCMMetadataTest(SolventMetadataTest):
    """Check we can parse implicit solvent data."""

    model = "SMD-CPCM"


class QChemSMDIEFPCMMetadataTest(QChemSolventMetadataTest, SMDIEFPCMMetadataTest):
    """Check we can parse implicit solvent data."""


class QChemSMDCPCMMetadataTest(QChemSolventMetadataTest, SMDCPCMMetadataTest):
    """Check we can parse implicit solvent data."""


if __name__ == "__main__":

    import sys
    sys.path.insert(1, os.path.join(__filedir__, ".."))

    from test_data import DataSuite
    suite = DataSuite(['SP'])
    suite.testall()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

"""Test the various population analyses (MPA, LPA, CSPA, Bickelhaupt) in cclib"""

import sys
import os
import logging
import unittest
from typing import Type

import numpy

from cclib.method import CSPA, LPA, MPA, OPA, Bickelhaupt
from cclib.method.calculationmethod import MissingAttributeError, Method
from cclib.parser import Gaussian

sys.path.insert(1, "..")

from ..test_data import getdatafile
import pytest


class PopulationTest(unittest.TestCase):
    """Generic population method tests."""
    
    methods = (CSPA, LPA, MPA, OPA, Bickelhaupt)

    def parse(self) -> None:
        self.data, self.logfile = getdatafile(Gaussian, "basicGaussian09", ["dvb_un_sp.log"])

    def calculate(self, method_class: Type[Method]) -> None:
        if not hasattr(self, 'data'):
            self.parse()
        self.analysis = method_class(self.data)
        self.analysis.logger.setLevel(0)
        self.analysis.calculate()

    def testmissingrequiredattributes(self) -> None:
        """Is an error raised when required attributes are missing?"""
        for missing_attribute in MPA.required_attrs:
            self.parse()
            delattr(self.data, missing_attribute)
            for method_class in self.methods:
                with pytest.raises(MissingAttributeError):
                    self.calculate(method_class)

    def testmissingoverlaps(self) -> None:
        """Is an error raised when no overlaps are available?"""
        self.parse()
        for overlap_attribute in MPA.overlap_attributes:
            if hasattr(self.data, overlap_attribute):
                delattr(self.data, overlap_attribute)
        for method_class in self.methods:
            if method_class.overlap_attributes:
                with pytest.raises(MissingAttributeError):
                    self.calculate(method_class)


class GaussianMPATest(unittest.TestCase):
    """Mulliken Population Analysis test"""

    def setUp(self) -> None:
        self.data, self.logfile = getdatafile(Gaussian, "basicGaussian09", ["dvb_un_sp.log"])
        self.analysis = MPA(self.data)
        self.analysis.logger.setLevel(0)
        self.analysis.calculate()

    def testsumcharges(self) -> None:
        """Do the Mulliken charges sum up to the total formal charge?"""
        formalcharge = sum(self.data.atomnos) - self.data.charge
        totalpopulation = sum(self.analysis.fragcharges)
        assert abs(totalpopulation-formalcharge) < 1.0e-3

    def testsumspins(self) -> None:
        """Do the Mulliken spins sum up to the total formal spin?"""
        formalspin = self.data.homos[0] - self.data.homos[1]
        totalspin = sum(self.analysis.fragspins)
        assert abs(totalspin-formalspin) < 1.0e-3


class GaussianLPATest(unittest.TestCase):
    """Lowdin Population Analysis test"""

    def setUp(self) -> None:
        self.data, self.logfile = getdatafile(Gaussian, "basicGaussian09", ["dvb_un_sp.log"])
        self.analysis = LPA(self.data)
        self.analysis.logger.setLevel(0)
        self.analysis.calculate()

    def testsumcharges(self) -> None:
        """Do the Lowdin charges sum up to the total formal charge?"""
        formalcharge = sum(self.data.atomnos) - self.data.charge
        totalpopulation = sum(self.analysis.fragcharges)
        assert abs(totalpopulation-formalcharge) < 0.001

    def testsumspins(self) -> None:
        """Do the Lowdin spins sum up to the total formal spin?"""
        formalspin = self.data.homos[0] - self.data.homos[1]
        totalspin = sum(self.analysis.fragspins)
        assert abs(totalspin-formalspin) < 1.0e-3


class GaussianCSPATest(unittest.TestCase):
    """C-squared Population Analysis test"""

    def setUp(self) -> None:
        self.data, self.logfile = getdatafile(Gaussian, "basicGaussian09", ["dvb_un_sp.log"])
        self.analysis = CSPA(self.data)
        self.analysis.logger.setLevel(0)
        self.analysis.calculate()

    def testsumcharges(self) -> None:
        """Do the CSPA charges sum up to the total formal charge?"""
        formalcharge = sum(self.data.atomnos) - self.data.charge
        totalpopulation = sum(self.analysis.fragcharges)
        assert abs(totalpopulation-formalcharge) < 1.0e-3

    def testsumspins(self) -> None:
        """Do the CSPA spins sum up to the total formal spin?"""
        formalspin = self.data.homos[0] - self.data.homos[1]
        totalspin = sum(self.analysis.fragspins)
        assert abs(totalspin-formalspin) < 1.0e-3

class GaussianBickelhauptTest(unittest.TestCase):
    """Bickelhaupt Population Analysis test"""
    
    def setUp(self) -> None:
        super(GaussianBickelhauptTest, self).setUp()
        self.data, self.logfile = getdatafile(Gaussian, "basicGaussian09", ["dvb_un_sp.log"])
        self.analysis = Bickelhaupt(self.data)
        self.analysis.logger.setLevel(0)
        self.analysis.calculate()
    
    def testsumcharges(self) -> None:
        """Do the Bickelhaupt charges sum up to the total formal charge?"""
        formalcharge = sum(self.data.atomnos) - self.data.charge
        totalpopulation = sum(self.analysis.fragcharges)
        assert abs(totalpopulation-formalcharge) < 1.0e-3
        
    def testsumspins(self) -> None:
        """Do the Bickelhaupt spins sum up to the total formal spin?"""
        formalspin = self.data.homos[0] - self.data.homos[1]
        totalspin = sum(self.analysis.fragspins)
        assert abs(totalspin-formalspin) < 1.0e-3
    
    def test_dvb_sp(self) -> None:
        """Testing Bickelhaupt charges (restricted) against outputs from Multiwfn."""
        data, logfile = getdatafile(Gaussian, "basicGaussian09", ["dvb_sp.out"])
        bpa = Bickelhaupt(data)
        bpa.logger.setLevel(logging.ERROR)
        bpa.calculate()
        
        e_bpa = numpy.loadtxt(f"{os.path.dirname(os.path.realpath(__file__))}/dvb_sp.bpa")
        assert numpy.all(bpa.fragcharges >= e_bpa - 0.05)
        assert numpy.all(bpa.fragcharges <= e_bpa + 0.05)
        
    def test_dvb_un_sp(self) -> None:
        """Testing Bickelhaupt charges (unrestricted) against outputs from Multiwfn."""
        data, logfile = getdatafile(Gaussian, "basicGaussian09", ["dvb_un_sp.log"])
        bpa = Bickelhaupt(data)
        bpa.logger.setLevel(logging.ERROR)
        bpa.calculate()
        
        e_bpaalpha = numpy.loadtxt(f"{os.path.dirname(os.path.realpath(__file__))}/dvb_un_sp.bpa")
        e_bpaspin = numpy.loadtxt(f"{os.path.dirname(os.path.realpath(__file__))}/dvb_un_sp.bpaspin")
        
        assert numpy.all(bpa.fragcharges >= e_bpaalpha - 0.05)
        assert numpy.all(bpa.fragcharges <= e_bpaalpha + 0.05)
        assert numpy.all(bpa.fragspins >= e_bpaspin - 0.05)
        assert numpy.all(bpa.fragspins <= e_bpaspin + 0.05)

tests = [GaussianMPATest, GaussianLPATest, GaussianCSPATest, GaussianBickelhauptTest]


if __name__ == "__main__":
    for test in tests:
        thistest = unittest.makeSuite(test)
        unittest.TextTestRunner(verbosity=2).run(thistest)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

"""Test the various population analyses (MPA, LPA, CSPA, Bickelhaupt) in cclib"""

import os
import unittest

import numpy

from cclib.method import Bader, volume
from cclib.parser import Psi4
from cclib.io import ccread
from cclib.method.calculationmethod import MissingAttributeError

from numpy.testing import assert_allclose

from ..test_data import getdatafile
import pytest


class BaderTest(unittest.TestCase):
    """Bader's QTAIM method tests."""

    def setUp(self) -> None:
        super(BaderTest, self).setUp()
        self.parse()

    def parse(self) -> None:
        self.data, self.logfile = getdatafile(Psi4, "basicPsi4-1.2.1", ["water_mp2.out"])
        self.volume = volume.Volume((-4, -4, -4), (4, 4, 4), (0.2, 0.2, 0.2))

    def testmissingrequiredattributes(self) -> None:
        """Is an error raised when required attributes are missing?"""
        for missing_attribute in Bader.required_attrs:
            self.parse()
            delattr(self.data, missing_attribute)
            with pytest.raises(MissingAttributeError):
                trialBader = Bader(self.data, self.volume)

    def test_val(self) -> None:
        """Do the calculated values match with known values?
        """

        self.data, logfile = getdatafile(Psi4, "basicPsi4-1.2.1", ["water_mp2.out"])
        self.volume = volume.Volume((-4, -4, -4), (4, 4, 4), (0.2, 0.2, 0.2))
        self.analysis = Bader(self.data, self.volume)
        self.analysis.calculate()

        refData = [6.9378, 0.3639, 0.3827]  # values from `bader` package

        assert_allclose(self.analysis.fragcharges, refData, atol=0.15)

    def test_chgsum_hf(self) -> None:
        """Does the sum of charges equate to the number of electrons for a simple molecule?"""

        hfpath = os.path.join(os.path.dirname(os.path.realpath(__file__)), "hf.out")
        data = ccread(hfpath)
        vol = volume.Volume((-6, -6, -6), (6, 6, 6), (0.2, 0.2, 0.2))
        analysis = Bader(data, vol)
        analysis.calculate()

        assert abs(numpy.sum(analysis.fragcharges)-10) < 1

    def test_symms_benzene(self) -> None:
        """ Do the carbons in benzene ring get assigned with roughly equal charges?
        
            Discrepancy between carbons do exist in this test due to grid coarseness and limited
            size of the grid. One can do a larger test, for example, using 160x170x80 size grid to
            obtain [5.63728706, 5.89862956, 5.73956182, 5.63728706, 5.73963179, 5.8996759].
            In comparison, `bader`, which is implemented by Henkelman group which proposed
            the algorithm, reports [5.947370, 6.032509, 5.873431, 5.947370, 5.873431, 6.033485].
            `bader` uses fuzzy boundaries which result in slightly higher carbon charges.
        """

        benzenepath = os.path.join(os.path.dirname(os.path.realpath(__file__)), "benzene.out")
        data = ccread(benzenepath)
        vol = volume.read_from_cube(
            os.path.join(os.path.dirname(os.path.realpath(__file__)), "benzene.cube")
        )
        assert_allclose(vol.origin, numpy.array([-4.1805  , -4.498006, -2.116709]), atol=1e-3)

        analysis = Bader(data, vol)
        analysis.calculate()

        assert abs(analysis.fragcharges[0:6].max()-analysis.fragcharges[0:6].min()) < 0.5



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

import unittest

import numpy as np

from cclib.bridge import cclib2pyscf
from cclib.parser.utils import find_package, convertor

from test.test_data import getdatafile


class PyscfTest(unittest.TestCase):

    def setUp(self) -> None:
        super(PyscfTest, self).setUp()
        if not find_package("pyscf"):
            raise ImportError("Must install pyscf to run this test")
        self.data, self.logfile = getdatafile(
            "GAMESS", "basicGAMESS-US2018", ["dvb_sp.out"]
        )
        self.udata, self.ulogfile = getdatafile(
            "GAMESS", "basicGAMESS-US2018", ["dvb_un_sp.out"]
        )

    def test_makepyscf(self) -> None:
        import pyscf
        from pyscf import dft

        refen = convertor(self.data.scfenergies[-1],"eV","hartree")  # value in eVs
        pyscfmol = cclib2pyscf.makepyscf(self.data)
        pyscfmol.cart = True
        pyscfmol.verbose = 0
        pyscfmol.build()

        mhf = dft.RKS(pyscfmol)
        mhf.xc = 'b3lyp'
        en = mhf.kernel()
        assert abs(en - refen) < 1.0e-4
        # check that default basis is returned if basis is not present.
        del self.data.gbasis
        pyscfmol2 = cclib2pyscf.makepyscf(self.data)
        assert pyscfmol2.basis == "sto-3g"

    def test_makepyscf_mos(self) -> None:
        pyscfmol = cclib2pyscf.makepyscf(self.data)
        mo_coeff, mo_occ, mo_syms, mo_energies = cclib2pyscf.makepyscf_mos(self.data,pyscfmol)
        assert np.allclose(mo_energies,convertor(np.array(self.data.moenergies),"eV","hartree"))
        # check first MO coefficient
        assert np.allclose(mo_coeff[0][0], self.data.mocoeffs[0][0][0])
        # check a random middle MO coefficient
        assert np.allclose(mo_coeff[0][10],self.data.mocoeffs[0][10][0])
        # test unrestricted code.
        pyscfmol = cclib2pyscf.makepyscf(self.udata)
        mo_coeff, mo_occ, mo_syms, mo_energies = cclib2pyscf.makepyscf_mos(self.udata,pyscfmol)
        assert np.allclose(mo_energies,convertor(np.array(self.udata.moenergies),"eV","hartree"))
        # check first MO coefficient
        assert np.allclose(mo_coeff[0][0][0], self.udata.mocoeffs[0][0][0])
        # check a random middle MO coefficient
        assert np.allclose(mo_coeff[0][0][10],self.udata.mocoeffs[0][10][0])

if __name__ == "__main__":
    unittest.main()
    PyscfTest.test_makepyscf_from_mos()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

"""Calculate properties of nuclei based on data parsed by cclib."""

import logging
from typing import Tuple

import numpy as np

from cclib.method.calculationmethod import Method
from cclib.parser.utils import PeriodicTable
from cclib.parser.utils import find_package, convertor

_found_periodictable = find_package("periodictable")
if _found_periodictable:
    import periodictable as pt

_found_scipy = find_package("scipy")
if _found_scipy:
    import scipy.constants


def _check_periodictable(found_periodictable: bool) -> None:
    if not _found_periodictable:
        raise ImportError("You must install `periodictable` to use this function")


def _check_scipy(found_scipy: bool) -> None:
    if not _found_scipy:
        raise ImportError("You must install `scipy` to use this function")


def get_most_abundant_isotope(element: "pt.Element") -> "pt.Isotope":
    """Given a `periodictable` element, return the most abundant
    isotope.
    """
    most_abundant_isotope = element.isotopes[0]
    abundance = 0
    for iso in element:
        if iso.abundance > abundance:
            most_abundant_isotope = iso
            abundance = iso.abundance
    return most_abundant_isotope


def get_isotopic_masses(charges):
    """Return the masses for the given nuclei, respresented by their
    nuclear charges.
    """
    _check_periodictable(_found_periodictable)
    masses = []
    for charge in charges:
        el = pt.elements[charge]
        isotope = get_most_abundant_isotope(el)
        mass = isotope.mass
        masses.append(mass)
    return np.array(masses)


class Nuclear(Method):
    """A container for methods pertaining to atomic nuclei."""

    def __init__(self, data, progress=None, loglevel=logging.INFO, logname="Log") -> None:
        super().__init__(data, progress, loglevel, logname)
        self.required_attrs = ('natom','atomcoords','atomnos','charge')

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return "Nuclear"

    def __repr__(self) -> str:
        """Return a representation of the object."""
        return "Nuclear"

    def stoichiometry(self) -> str:
        """Return the stoichemistry of the object according to the Hill system"""
        cclib_pt = PeriodicTable()
        elements = [cclib_pt.element[ano] for ano in self.data.atomnos]
        counts = {el: elements.count(el) for el in set(elements)}

        formula = ""
        elcount = lambda el, c: f"{el}{int(c)}" if c > 1 else el
        if 'C' in elements:
            formula += elcount('C', counts['C'])
            counts.pop('C')
            if 'H' in elements:
              formula += elcount('H', counts['H'])
              counts.pop('H')
        for el, c in sorted(counts.items()):
            formula += elcount(el, c)

        if getattr(self.data, 'charge', 0):
            magnitude = abs(self.data.charge)
            sign = "+" if self.data.charge > 0 else "-"
            formula += f"({sign}{int(magnitude)})"
        return formula

    def repulsion_energy(self, atomcoords_index: int = -1) -> float:
        """Return the nuclear repulsion energy."""
        nre = 0.0
        atomcoords = convertor(self.data.atomcoords[atomcoords_index], "Angstrom", "bohr")
        for i in range(self.data.natom):
            ri = atomcoords[i]
            zi = self.data.atomnos[i]
            for j in range(i+1, self.data.natom):
                rj = atomcoords[j]
                zj = self.data.atomnos[j]
                d = np.linalg.norm(ri-rj)
                nre += zi * zj / d
        return convertor(nre, "hartree", "eV")

    def center_of_mass(self, atomcoords_index: int = -1) -> float:
        """Return the center of mass."""
        charges = self.data.atomnos
        coords = self.data.atomcoords[atomcoords_index]
        masses = get_isotopic_masses(charges)

        mwc = coords * masses[:, np.newaxis]
        numerator = np.sum(mwc, axis=0)
        denominator = np.sum(masses)

        return numerator / denominator

    def moment_of_inertia_tensor(self, atomcoords_index: int = -1) -> np.ndarray:
        """Return the moment of inertia tensor."""
        charges = self.data.atomnos
        coords = self.data.atomcoords[atomcoords_index]
        masses = get_isotopic_masses(charges)

        moi_tensor = np.empty((3, 3))

        moi_tensor[0][0] = np.sum(masses * (coords[:, 1]**2 + coords[:, 2]**2))
        moi_tensor[1][1] = np.sum(masses * (coords[:, 0]**2 + coords[:, 2]**2))
        moi_tensor[2][2] = np.sum(masses * (coords[:, 0]**2 + coords[:, 1]**2))

        moi_tensor[0][1] = np.sum(masses * coords[:, 0] * coords[:, 1])
        moi_tensor[0][2] = np.sum(masses * coords[:, 0] * coords[:, 2])
        moi_tensor[1][2] = np.sum(masses * coords[:, 1] * coords[:, 2])

        moi_tensor[1][0] = moi_tensor[0][1]
        moi_tensor[2][0] = moi_tensor[0][2]
        moi_tensor[2][1] = moi_tensor[1][2]

        return moi_tensor

    def principal_moments_of_inertia(self, units: str = 'amu_bohr_2') -> Tuple[np.ndarray, np.ndarray]:
        """Return the principal moments of inertia in 3 kinds of units:
        1. [amu][bohr]^2
        2. [amu][angstrom]^2
        3. [g][cm]^2
        and the principal axes.
        """
        choices = ('amu_bohr_2', 'amu_angstrom_2', 'g_cm_2')
        units = units.lower()
        if units not in choices:
            raise ValueError(f"Invalid units, pick one of {choices}")
        moi_tensor = self.moment_of_inertia_tensor()
        principal_moments, principal_axes = np.linalg.eigh(moi_tensor)
        if units == "amu_bohr_2":
            _check_scipy(_found_scipy)
            bohr2ang = scipy.constants.value("atomic unit of length") / scipy.constants.angstrom
            conv = 1 / bohr2ang ** 2
        elif units == "amu_angstrom_2":
            conv = 1
        elif units == "g_cm_2":
            _check_scipy(_found_scipy)
            amu2g = scipy.constants.value("unified atomic mass unit") * scipy.constants.kilo
            conv = amu2g * (scipy.constants.angstrom / scipy.constants.centi) ** 2
        return conv * principal_moments, principal_axes

    def rotational_constants(self, units: str = 'ghz') -> np.ndarray:
        """Compute the rotational constants in 1/cm or GHz."""
        choices = ('invcm', 'ghz')
        units = units.lower()
        if units not in choices:
            raise ValueError(f"Invalid units, pick one of {choices}")
        principal_moments = self.principal_moments_of_inertia("amu_angstrom_2")[0]
        _check_scipy(_found_scipy)
        bohr2ang = scipy.constants.value('atomic unit of length') / scipy.constants.angstrom
        xfamu = 1 / scipy.constants.value('electron mass in u')
        xthz = scipy.constants.value('hartree-hertz relationship')
        rotghz = xthz * (bohr2ang ** 2) / (2 * xfamu * scipy.constants.giga)
        if units == 'ghz':
            conv = rotghz
        if units == 'invcm':
            ghz2invcm = scipy.constants.giga * scipy.constants.centi / scipy.constants.c
            conv = rotghz * ghz2invcm
        return conv / principal_moments


del find_package



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

"""Parser for Molcas output files"""

import re
import string

import numpy

from cclib.parser import logfileparser
from cclib.parser import utils


class Molcas(logfileparser.Logfile):
    """A Molcas log file."""

    def __init__(self, *args, **kwargs):
        super().__init__(logname="Molcas", *args, **kwargs)

    def __str__(self):
        """Return a string repeesentation of the object."""
        return f"Molcas log file {self.filename}"

    def __repr__(self):
        """Return a representation of the object."""
        return f'Molcas("{self.filename}")'

    def normalisesym(self, label):
        """Normalise the symmetries used by Molcas.

        The labels are standardized except for the first character being lowercase.
        """
        return label[0].upper() + label[1:]

    def after_parsing(self):
        for element, ncore in self.core_array:
            self._assign_coreelectrons_to_element(element, ncore)

        if "package_version" in self.metadata:
            # Use the short version as the legacy version.
            self.metadata["legacy_package_version"] = self.metadata["package_version"]
            # If there is both a tag and the full hash, place the tag
            # first. Both are chosen to be local, since there isn't a
            # distinction between development and release builds in their
            # version cycle.
            if "tag" in self.metadata and "revision" in self.metadata:
                self.metadata[
                    "package_version"
                ] = f"{self.metadata['package_version']}+{self.metadata['tag']}.{self.metadata['revision']}"
            elif "tag" in self.metadata:
                self.metadata[
                    "package_version"
                ] = f"{self.metadata['package_version']}+{self.metadata['tag']}"
            elif "revision" in self.metadata:
                self.metadata[
                    "package_version"
                ] = f"{self.metadata['package_version']}+{self.metadata['revision']}"

    def before_parsing(self):
        # Compile the regex for extracting the element symbol from the
        # atom label in the "Molecular structure info" block.
        self.re_atomelement = re.compile(r'([a-zA-Z]+)\d?')

        # Compile the dashes-and-or-spaces-only regex.
        self.re_dashes_and_spaces = re.compile(r'^[\s-]+$')

        # Molcas can do multiple calculations in one job, and each one
        # starts from the gateway module. Onle parse the first.
        # TODO: It would be best to parse each calculation as a separate
        # ccData object and return an iterator - something for 2.x
        self.gateway_module_count = 0

    def extract(self, inputfile, line):
        """Extract information from the file object inputfile."""

        if "Start Module: gateway" in line:
            self.gateway_module_count += 1

        if self.gateway_module_count > 1:
            return

        # Extract the version number and optionally the Git tag and hash.
        if "version" in line:
            match = re.search(r"\s{2,}version:?\s(\d*\.\d*)", line)
            if match:
                self.metadata["package_version"] = match.groups()[0]
        if "tag" in line:
            self.metadata["tag"] = line.split()[-1]
        if "build" in line:
            match = re.search(r"\*\s*build\s(\S*)\s*\*", line)
            if match:
                self.metadata["revision"] = match.groups()[0]

        ## This section is present when executing &GATEWAY.
        # ++    Molecular structure info:
        #       -------------------------

        #                     ************************************************
        #                     **** Cartesian Coordinates / Bohr, Angstrom ****
        #                     ************************************************

        #      Center  Label                x              y              z                     x              y              z
        #         1      C1               0.526628      -2.582937       0.000000              0.278679      -1.366832       0.000000
        #         2      C2               2.500165      -0.834760       0.000000              1.323030      -0.441736       0.000000
        if line[25:63] == 'Cartesian Coordinates / Bohr, Angstrom':
            if not hasattr(self, 'atomnos'):
                self.atomnos = []

            self.skip_lines(inputfile, ['stars', 'blank', 'header'])

            line = next(inputfile)

            atomelements = []
            atomcoords = []

            while line.strip() not in ('', '--'):
                sline = line.split()
                atomelement = sline[1].rstrip(string.digits).title()
                atomelements.append(atomelement)
                atomcoords.append(list(map(float, sline[5:])))
                line = next(inputfile)

            self.append_attribute('atomcoords', atomcoords)

            if self.atomnos == []:
                self.atomnos = [self.table.number[ae.title()] for ae in atomelements]

            if not hasattr(self, 'natom'):
                self.set_attribute('natom', len(self.atomnos))

        ## This section is present when executing &SCF.
        #  ++    Orbital specifications:
        #  -----------------------

        #  Symmetry species               1

        #  Frozen orbitals                0
        #  Occupied orbitals              3
        #  Secondary orbitals            77
        #  Deleted orbitals               0
        #  Total number of orbitals      80
        #  Number of basis functions     80
        #  --
        if line[:29] == '++    Orbital specifications:':

            self.skip_lines(inputfile, ['dashes', 'blank'])
            line = next(inputfile)

            symmetry_count = 1
            while not line.startswith('--'):
                if line.strip().startswith('Symmetry species'):
                    symmetry_count = int(line.split()[-1])
                if line.strip().startswith('Total number of orbitals'):
                    nmos = line.split()[-symmetry_count:]
                    self.set_attribute('nmo', sum(map(int, nmos)))
                if line.strip().startswith('Number of basis functions'):
                    nbasis = line.split()[-symmetry_count:]
                    self.set_attribute('nbasis', sum(map(int, nbasis)))

                line = next(inputfile)

        if line.strip().startswith(('Molecular charge', 'Total molecular charge')):
            self.set_attribute('charge', int(float(line.split()[-1])))

        #  ++    Molecular charges:
        #  ------------------

        #  Mulliken charges per centre and basis function type
        #  ---------------------------------------------------

        #         C1
        #  1s     2.0005
        #  2s     2.0207
        #  2px    0.0253
        #  2pz    0.1147
        #  2py    1.8198
        #  *s    -0.0215
        #  *px    0.0005
        #  *pz    0.0023
        #  *py    0.0368
        #  *d2+   0.0002
        #  *d1+   0.0000
        #  *d0    0.0000
        #  *d1-   0.0000
        #  *d2-   0.0000
        #  *f3+   0.0000
        #  *f2+   0.0001
        #  *f1+   0.0000
        #  *f0    0.0001
        #  *f1-   0.0001
        #  *f2-   0.0000
        #  *f3-   0.0003
        #  *g4+   0.0000
        #  *g3+   0.0000
        #  *g2+   0.0000
        #  *g1+   0.0000
        #  *g0    0.0000
        #  *g1-   0.0000
        #  *g2-   0.0000
        #  *g3-   0.0000
        #  *g4-   0.0000
        #  Total  6.0000

        #  N-E    0.0000

        #  Total electronic charge=    6.000000

        #  Total            charge=    0.000000
        #--
        if line[:24] == '++    Molecular charges:':

            atomcharges = []

            while line[6:29] != 'Total electronic charge':
                line = next(inputfile)
                if line[6:9] == 'N-E':
                    atomcharges.extend(map(float, line.split()[1:]))

            # Molcas only performs Mulliken population analysis.
            self.set_attribute('atomcharges', {'mulliken': atomcharges})

            # Ensure the charge printed here is identical to the
            # charge printed before entering the SCF.
            self.skip_line(inputfile, 'blank')
            line = next(inputfile)
            assert line[6:30] == 'Total            charge='
            if hasattr(self, 'charge'):
                assert int(float(line.split()[2])) == self.charge

        # This section is present when executing &SCF
        # This section parses the total SCF Energy.
        # *****************************************************************************************************************************
        # *                                                                                                                           *
        # *                                             SCF/KS-DFT Program, Final results                                             *
        # *                                                                                                                           *
        # *                                                                                                                           *
        # *                                                                                                                           *
        # *                                                       Final Results                                                       *
        # *                                                                                                                           *
        # *****************************************************************************************************************************

        # ::    Total SCF energy                                -37.6045426484
        if line[:22] == '::    Total SCF energy' or line[:25] == '::    Total KS-DFT energy':
            if not hasattr(self, 'scfenergies'):
                self.scfenergies = []
            scfenergy = float(line.split()[-1])
            self.scfenergies.append(utils.convertor(scfenergy, 'hartree', 'eV'))

        ## Parsing the scftargets in this section
        #  ++    Optimization specifications:
        #  ----------------------------

        #  SCF Algorithm: Conventional
        #  Minimized density differences are used
        #  Number of density matrices in core                9
        #  Maximum number of NDDO SCF iterations           400
        #  Maximum number of HF  SCF iterations            400
        #  Threshold for SCF energy change            0.10E-08
        #  Threshold for density matrix               0.10E-03
        #  Threshold for Fock matrix                  0.15E-03
        #  Threshold for linear dependence            0.10E-08
        #  Threshold at which DIIS is turned on       0.15E+00
        #  Threshold at which QNR/C2DIIS is turned on 0.75E-01
        #  Threshold for Norm(delta) (QNR/C2DIIS)     0.20E-04
        if line[:34] == '++    Optimization specifications:':
            self.skip_lines(inputfile, ['d', 'b'])
            line = next(inputfile)
            if line.strip().startswith('SCF'):
                scftargets = []
                self.skip_lines(inputfile,
                                ['Minimized', 'Number', 'Maximum', 'Maximum'])
                lines = [next(inputfile) for i in range(7)]
                targets = [
                    'Threshold for SCF energy change',
                    'Threshold for density matrix',
                    'Threshold for Fock matrix',
                    'Threshold for Norm(delta)',
                ]
                for y in targets:
                    scftargets.extend([float(x.split()[-1]) for x in lines if y in x])

                self.append_attribute('scftargets', scftargets)

        #  ++ Convergence information
        #                                     SCF        iterations: Energy and convergence statistics
        #
        #  Iter     Tot. SCF       One-electron     Two-electron   Energy   Max Dij or  Max Fij    DNorm      TNorm     AccCon    Time
        #             Energy          Energy          Energy       Change   Delta Norm                                          in Sec.
        #     1    -36.83817703    -50.43096166     13.59278464  0.00E+00   0.16E+00*  0.27E+01*   0.30E+01   0.33E+02   NoneDa    0.
        #     2    -36.03405202    -45.74525152      9.71119950  0.80E+00*  0.14E+00*  0.93E-02*   0.26E+01   0.43E+01   Damp      0.
        #     3    -37.08936118    -48.41536598     11.32600480 -0.11E+01*  0.12E+00*  0.91E-01*   0.97E+00   0.16E+01   Damp      0.
        #     4    -37.31610460    -50.54103969     13.22493509 -0.23E+00*  0.11E+00*  0.96E-01*   0.72E+00   0.27E+01   Damp      0.
        #     5    -37.33596239    -49.47021484     12.13425245 -0.20E-01*  0.59E-01*  0.59E-01*   0.37E+00   0.16E+01   Damp      0.
        # ...
        #           Convergence after 26 Macro Iterations
        # --
        if line[46:91] == 'iterations: Energy and convergence statistics':

            self.skip_line(inputfile, 'blank')

            while line.split() != ['Energy', 'Energy', 'Energy', 'Change', 'Delta', 'Norm', 'in', 'Sec.']:
                line = next(inputfile)

            iteration_regex = (r"^([0-9]+)"                                  # Iter
                               r"( [ \-0-9]*\.[0-9]{6,9})"                   # Tot. SCF Energy
                               r"( [ \-0-9]*\.[0-9]{6,9})"                   # One-electron Energy
                               r"( [ \-0-9]*\.[0-9]{6,9})"                   # Two-electron Energy
                               r"( [ \-0-9]*\.[0-9]{2}E[\-\+][0-9]{2}\*?)"   # Energy Change
                               r"( [ \-0-9]*\.[0-9]{2}E[\-\+][0-9]{2}\*?)"   # Max Dij or Delta Norm
                               r"( [ \-0-9]*\.[0-9]{2}E[\-\+][0-9]{2}\*?)"   # Max Fij
                               r"( [ \-0-9]*\.[0-9]{2}E[\-\+][0-9]{2}\*?)"   # DNorm
                               r"( [ \-0-9]*\.[0-9]{2}E[\-\+][0-9]{2}\*?)"   # TNorm
                               r"( [ A-Za-z0-9]*)"                           # AccCon
                               r"( [ \.0-9]*)$")                             # Time in Sec.

            scfvalues = []
            line = next(inputfile)
            while not line.strip().startswith("Convergence"):

                match = re.match(iteration_regex, line.strip())
                if match:
                    groups = match.groups()
                    cols = [g.strip() for g in match.groups()]
                    cols = [c.replace('*', '') for c in cols]

                    energy = float(cols[4])
                    density = float(cols[5])
                    fock = float(cols[6])
                    dnorm = float(cols[7])
                    scfvalues.append([energy, density, fock, dnorm])

                if line.strip() == "--":
                    self.logger.warning('File terminated before end of last SCF!')
                    break

                line = next(inputfile)

            self.append_attribute('scfvalues', scfvalues)

        #  Harmonic frequencies in cm-1
        #
        #  IR Intensities in km/mol
        #
        #                         1         2         3         4         5         6
        #
        #      Frequency:       i60.14    i57.39    128.18    210.06    298.24    309.65
        #
        #      Intensity:    3.177E-03 2.129E-06 4.767E-01 2.056E-01 6.983E-07 1.753E-07
        #      Red. mass:      2.42030   2.34024   2.68044   3.66414   2.61721   3.34904
        #
        #      C1         x   -0.00000   0.00000   0.00000  -0.05921   0.00000  -0.06807
        #      C1         y    0.00001  -0.00001  -0.00001   0.00889   0.00001  -0.02479
        #      C1         z   -0.03190   0.04096  -0.03872   0.00001  -0.12398  -0.00002
        #      C2         x   -0.00000   0.00001   0.00000  -0.06504   0.00000  -0.03487
        #      C2         y    0.00000  -0.00000  -0.00000   0.01045   0.00001  -0.05659
        #      C2         z   -0.03703  -0.03449  -0.07269   0.00000  -0.07416  -0.00001
        #      C3         x   -0.00000   0.00001   0.00000  -0.06409  -0.00001   0.05110
        #      C3         y   -0.00000   0.00001   0.00000   0.00152   0.00000  -0.03263
        #      C3         z   -0.03808  -0.08037  -0.07267  -0.00001   0.07305   0.00000
        # ...
        #      H20        y    0.00245  -0.00394   0.03215   0.03444  -0.10424  -0.10517
        #      H20        z    0.00002  -0.00001   0.00000  -0.00000  -0.00000   0.00000
        #
        #
        #
        # ++ Thermochemistry
        if line[1:29] == 'Harmonic frequencies in cm-1':

            self.skip_line(inputfile, 'blank')
            line = next(inputfile)

            while 'Thermochemistry' not in line:

                if 'Frequency:' in line:
                    if not hasattr(self, 'vibfreqs'):
                        self.vibfreqs = []
                    vibfreqs = [float(i.replace('i', '-')) for i in line.split()[1:]]
                    self.vibfreqs.extend(vibfreqs)

                if 'Intensity:' in line:
                    if not hasattr(self, 'vibirs'):
                        self.vibirs = []
                    vibirs = map(float, line.split()[1:])
                    self.vibirs.extend(vibirs)

                if 'Red.' in line:
                    if not hasattr(self, 'vibrmasses'):
                        self.vibrmasses = []
                    vibrmasses = map(float, line.split()[2:])
                    self.vibrmasses.extend(vibrmasses)

                    self.skip_line(inputfile, 'blank')
                    line = next(inputfile)
                    if not hasattr(self, 'vibdisps'):
                        self.vibdisps = []
                    disps = []
                    for n in range(3*self.natom):
                        numbers = [float(s) for s in line[17:].split()]
                        # The atomindex should start at 0 instead of 1.
                        atomindex = int(re.search(r'\d+$', line.split()[0]).group()) - 1
                        numbermodes = len(numbers)
                        if len(disps) == 0:
                            # Appends empty array of the following
                            # dimensions (numbermodes, natom, 0) to disps.
                            for mode in range(numbermodes):
                                disps.append([[] for x in range(0, self.natom)])
                        for mode in range(numbermodes):
                            disps[mode][atomindex].append(numbers[mode])
                        line = next(inputfile)
                    self.vibdisps.extend(disps)

                line = next(inputfile)

        ## Parsing thermochemistry attributes here
        #  ++ Thermochemistry
        #
        #   *********************
        #   *                   *
        #   *  THERMOCHEMISTRY  *
        #   *                   *
        #   *********************
        #
        #   Mass-centered Coordinates (Angstrom):
        #   ***********************************************************
        # ...
        #   *****************************************************
        #   Temperature =     0.00 Kelvin, Pressure =   1.00 atm
        #   -----------------------------------------------------
        #   Molecular Partition Function and Molar Entropy:
        #                          q/V (M**-3)    S(kcal/mol*K)
        #   Electronic            0.100000D+01        0.000
        #   Translational         0.100000D+01        0.000
        #   Rotational            0.100000D+01        2.981
        #   Vibrational           0.100000D+01        0.000
        #   TOTAL                 0.100000D+01        2.981
        #
        #   Thermal contributions to INTERNAL ENERGY:
        #   Electronic           0.000 kcal/mol      0.000000 au.
        #   Translational        0.000 kcal/mol      0.000000 au.
        #   Rotational           0.000 kcal/mol      0.000000 au.
        #   Vibrational        111.885 kcal/mol      0.178300 au.
        #   TOTAL              111.885 kcal/mol      0.178300 au.
        #
        #   Thermal contributions to
        #   ENTHALPY           111.885 kcal/mol      0.178300 au.
        #   GIBBS FREE ENERGY  111.885 kcal/mol      0.178300 au.
        #
        #   Sum of energy and thermal contributions
        #   INTERNAL ENERGY                       -382.121931 au.
        #   ENTHALPY                              -382.121931 au.
        #   GIBBS FREE ENERGY                     -382.121931 au.
        #   -----------------------------------------------------
        # ...
        #   ENTHALPY                              -382.102619 au.
        #   GIBBS FREE ENERGY                     -382.179819 au.
        #   -----------------------------------------------------
        #  --
        #
        #  ++    Isotopic shifts:
        if line[4:19] == 'THERMOCHEMISTRY':

            while "ZPVE" not in line:
                line = next(inputfile)
            self.set_attribute("zpve", float(line.split()[3]))

            temperature_values = []
            pressure_values = []
            entropy_values = []
            internal_energy_values = []
            enthalpy_values = []
            free_energy_values = []

            while 'Isotopic' not in line:

                if line[1:12] == 'Temperature':
                    temperature_values.append(float(line.split()[2]))
                    pressure_values.append(float(line.split()[6]))

                if line[1:48] == 'Molecular Partition Function and Molar Entropy:':
                    while 'TOTAL' not in line:
                        line = next(inputfile)
                    # Molcas reports entropy values in kcal/mol*K but actually writes them in cal/mol*K
                    entropy_values.append(utils.convertor(float(line.split()[2]), 'kcal/mol', 'hartree') / 1000)

                if line[1:40] == 'Sum of energy and thermal contributions':
                    internal_energy_values.append(float(next(inputfile).split()[2]))
                    enthalpy_values.append(float(next(inputfile).split()[1]))
                    free_energy_values.append(float(next(inputfile).split()[3]))

                line = next(inputfile)
            # When calculations for more than one temperature value are
            # performed, the values corresponding to room temperature (298.15 K)
            # are returned and if no calculations are performed for 298.15 K, then
            # the values corresponding last temperature value are returned.
            index = -1
            if 298.15 in temperature_values:
                index = temperature_values.index(298.15)

            self.set_attribute('temperature', temperature_values[index])
            if len(temperature_values) > 1:
                self.logger.warning('More than 1 values of temperature found')

            self.set_attribute('pressure', pressure_values[index])
            if len(pressure_values) > 1:
                self.logger.warning('More than 1 values of pressure found')

            self.set_attribute('entropy', entropy_values[index])
            if len(entropy_values) > 1:
                self.logger.warning('More than 1 values of entropy found')

            self.set_attribute('enthalpy', enthalpy_values[index])
            if len(enthalpy_values) > 1:
                self.logger.warning('More than 1 values of enthalpy found')

            self.set_attribute('freeenergy', free_energy_values[index])
            if len(free_energy_values) > 1:
                self.logger.warning('More than 1 values of freeenergy found')

        ## Parsing Geometrical Optimization attributes in this section.
        #  ++       Slapaf input parameters:
        #  ------------------------
        #
        # Max iterations:                            2000
        # Convergence test a la Schlegel.
        # Convergence criterion on gradient/para.<=: 0.3E-03
        # Convergence criterion on step/parameter<=: 0.3E-03
        # Convergence criterion on energy change <=: 0.0E+00
        # Max change of an internal coordinate:     0.30E+00
        # ...
        # ...
        #  **********************************************************************************************************************
        #  *                                    Energy Statistics for Geometry Optimization                                     *
        #  **********************************************************************************************************************
        #                          Energy     Grad      Grad              Step                 Estimated   Geom       Hessian
        #  Iter      Energy       Change     Norm      Max    Element    Max     Element     Final Energy Update Update   Index
        #    1   -382.30023222  0.00000000 0.107221  0.039531 nrc047   0.085726  nrc047     -382.30533799 RS-RFO  None      0
        #    2   -382.30702964 -0.00679742 0.043573  0.014908 nrc001   0.068195  nrc001     -382.30871333 RS-RFO  BFGS      0
        #    3   -382.30805348 -0.00102384 0.014883  0.005458 nrc010  -0.020973  nrc001     -382.30822089 RS-RFO  BFGS      0
        # ...
        # ...
        #   18   -382.30823419 -0.00000136 0.001032  0.000100 nrc053   0.012319  nrc053     -382.30823452 RS-RFO  BFGS      0
        #   19   -382.30823198  0.00000221 0.001051 -0.000092 nrc054   0.066565  nrc053     -382.30823822 RS-RFO  BFGS      0
        #   20   -382.30820252  0.00002946 0.001132 -0.000167 nrc021  -0.064003  nrc053     -382.30823244 RS-RFO  BFGS      0
        #
        #         +----------------------------------+----------------------------------+
        #         +    Cartesian Displacements       +    Gradient in internals         +
        #         +  Value      Threshold Converged? +  Value      Threshold Converged? +
        #   +-----+----------------------------------+----------------------------------+
        #   + RMS + 5.7330E-02  1.2000E-03     No    + 1.6508E-04  3.0000E-04     Yes   +
        #   +-----+----------------------------------+----------------------------------+
        #   + Max + 1.2039E-01  1.8000E-03     No    + 1.6711E-04  4.5000E-04     Yes   +
        #   +-----+----------------------------------+----------------------------------+
        if 'Convergence criterion on energy change' in line:
            self.energy_threshold = float(line.split()[6])
            # If energy change threshold equals zero,
            # then energy change is not a criteria for convergence.
            if self.energy_threshold == 0:
                self.energy_threshold = numpy.inf

        if 'Energy Statistics for Geometry Optimization' in line:
            if not hasattr(self, 'geovalues'):
                self.geovalues = []

            self.skip_lines(inputfile, ['stars', 'header'])
            line = next(inputfile)
            assert 'Iter      Energy       Change     Norm' in line
            # A variable keeping track of ongoing iteration.
            iter_number = len(self.geovalues) + 1
            # Iterate till blank line.
            while line.split() != []:
                for i in range(iter_number):
                    line = next(inputfile)
                self.geovalues.append([float(line.split()[2])])
                line = next(inputfile)
            # Along with energy change, RMS and Max values of change in
            # Cartesian Diaplacement and Gradients are used as optimization
            # criteria.
            self.skip_lines(inputfile, ['border', 'header', 'header', 'border'])
            line = next(inputfile)
            assert '+ RMS +' in line
            line_rms = line.split()
            line = next(inputfile)
            line_max = next(inputfile).split()
            if not hasattr(self, 'geotargets'):
                # The attribute geotargets is an array consisting of the following
                # values: [Energy threshold, Max Gradient threshold, RMS Gradient threshold, \
                #          Max Displacements threshold, RMS Displacements threshold].
                max_gradient_threshold = float(line_max[8])
                rms_gradient_threshold = float(line_rms[8])
                max_displacement_threshold = float(line_max[4])
                rms_displacement_threshold = float(line_rms[4])
                self.geotargets = [self.energy_threshold, max_gradient_threshold, rms_gradient_threshold, max_displacement_threshold, rms_displacement_threshold]

            max_gradient_change = float(line_max[7])
            rms_gradient_change = float(line_rms[7])
            max_displacement_change = float(line_max[3])
            rms_displacement_change = float(line_rms[3])
            self.geovalues[iter_number - 1].extend([max_gradient_change, rms_gradient_change, max_displacement_change, rms_displacement_change])

        #   *********************************************************
        #   * Nuclear coordinates for the next iteration / Angstrom *
        #   *********************************************************
        #    ATOM              X               Y               Z
        #    C1               0.235560       -1.415847        0.012012
        #    C2               1.313797       -0.488199        0.015149
        #    C3               1.087050        0.895510        0.014200
        # ...
        # ...
        #    H19             -0.021327       -4.934915       -0.029355
        #    H20             -1.432030       -3.721047       -0.039835
        #
        #  --
        if 'Nuclear coordinates for the next iteration / Angstrom' in line:
            self.skip_lines(inputfile, ['s', 'header'])
            line = next(inputfile)

            atomcoords = []
            while line.split() != []:
                atomcoords.append([float(c) for c in line.split()[1:]])
                line = next(inputfile)

            if len(atomcoords) == self.natom:
                self.atomcoords.append(atomcoords)
            else:
                self.logger.warning(
                        f"Parsed coordinates not consistent with previous, skipping. This could be due to symmetry being turned on during the job. Length was {len(self.atomcoords[-1])}, now found {len(atomcoords)}. New coordinates: {str(atomcoords)}")

        #  **********************************************************************************************************************
        #  *                                    Energy Statistics for Geometry Optimization                                     *
        #  **********************************************************************************************************************
        #                         Energy     Grad      Grad              Step                 Estimated   Geom       Hessian
        #  Iter      Energy       Change     Norm      Max    Element    Max     Element     Final Energy Update Update   Index
        #    1   -382.30023222  0.00000000 0.107221  0.039531 nrc047   0.085726  nrc047     -382.30533799 RS-RFO  None      0
        # ...
        # ...
        #   23   -382.30823115 -0.00000089 0.001030  0.000088 nrc053   0.000955  nrc053     -382.30823118 RS-RFO  BFGS      0
        #
        #         +----------------------------------+----------------------------------+
        #         +    Cartesian Displacements       +    Gradient in internals         +
        #         +  Value      Threshold Converged? +  Value      Threshold Converged? +
        #   +-----+----------------------------------+----------------------------------+
        #   + RMS + 7.2395E-04  1.2000E-03     Yes   + 2.7516E-04  3.0000E-04     Yes   +
        #   +-----+----------------------------------+----------------------------------+
        #   + Max + 1.6918E-03  1.8000E-03     Yes   + 8.7768E-05  4.5000E-04     Yes   +
        #   +-----+----------------------------------+----------------------------------+
        #
        #   Geometry is converged in  23 iterations to a Minimum Structure
        if 'Geometry is converged' in line:
            if not hasattr(self, 'optdone'):
                self.optdone = []
            self.optdone.append(len(self.atomcoords))

        #   *********************************************************
        #   * Nuclear coordinates of the final structure / Angstrom *
        #   *********************************************************
        #    ATOM              X               Y               Z
        #    C1               0.235547       -1.415838        0.012193
        #    C2               1.313784       -0.488201        0.015297
        #    C3               1.087036        0.895508        0.014333
        # ...
        # ...
        #    H19             -0.021315       -4.934913       -0.029666
        #    H20             -1.431994       -3.721026       -0.041078
        if 'Nuclear coordinates of the final structure / Angstrom' in line:
            self.skip_lines(inputfile, ['s', 'header'])
            line = next(inputfile)

            atomcoords = []

            while line.split() != []:
                atomcoords.append([float(c) for c in line.split()[1:]])
                line = next(inputfile)

            if len(atomcoords) == self.natom:
                self.atomcoords.append(atomcoords)
            else:
                self.logger.error(
                        f'Number of atoms ({len(atomcoords)}) in parsed atom coordinates is smaller than previously ({int(self.natom)}), possibly due to symmetry. Ignoring these coordinates.')

        ## Parsing Molecular Gradients attributes in this section.
        # ()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()
        # 
        #                                               &ALASKA
        # 
        #                                    only a single process is used
        #                        available to each process: 2.0 GB of memory, 1 thread
        # ()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()
        # ...
        # ...
        #  **************************************************
        #  *                                                *
        #  *              Molecular gradients               *
        #  *                                                *
        #  **************************************************
        # 
        #   Irreducible representation: a  
        #  ---------------------------------------------------------
        #                      X             Y             Z        
        #  ---------------------------------------------------------
        #   C1               -0.00009983   -0.00003043    0.00001004
        #   ...
        #   H20              -0.00027629    0.00010546    0.00003317
        #  ---------------------------------------------------
        # WARNING: "Molecular gradients, after ESPF" is found for ESPF QM/MM calculations
        if "Molecular gradients " in line:

            if not hasattr(self, "grads"):
                self.grads = []

            self.skip_lines(inputfile, ['stars', 'stars', 'blank', 'header',
                                        'dashes', 'header', 'dashes'])

            grads = []
            line = next(inputfile)
            while len(line.split()) == 4:
                tmpgrads = list(map(float, line.split()[1:]))
                grads.append(tmpgrads)
                line = next(inputfile)

            self.append_attribute('grads', grads)

        # This code here works, but QM/MM gradients are printed after QM ones.
        # Maybe another attribute is needed to store them to have both.
        if "Molecular gradients, after ESPF" in line:

            self.skip_lines(inputfile, ['stars', 'stars', 'blank', 'header',
                                        'dashes', 'header', 'dashes'])

            grads = []
            line = next(inputfile)
            while len(line.split()) == 4:
                tmpgrads = list(map(float, line.split()[1:]))
                grads.append(tmpgrads)
                line = next(inputfile)

            self.grads[-1] = grads

        ###
        #        All orbitals with orbital energies smaller than  E(LUMO)+0.5 are printed
        #
        #  ++    Molecular orbitals:
        #        -------------------
        #
        #        Title: RKS-DFT orbitals
        #
        #        Molecular orbitals for symmetry species 1: a
        #
        #            Orbital        1         2         3         4         5         6         7         8         9        10
        #            Energy      -10.0179  -10.0179  -10.0075  -10.0075  -10.0066  -10.0066  -10.0056  -10.0055   -9.9919   -9.9919
        #            Occ. No.      2.0000    2.0000    2.0000    2.0000    2.0000    2.0000    2.0000    2.0000    2.0000    2.0000
        #
        #          1 C1    1s     -0.6990    0.6989    0.0342    0.0346    0.0264   -0.0145   -0.0124   -0.0275   -0.0004   -0.0004
        #          2 C1    2s     -0.0319    0.0317   -0.0034   -0.0033   -0.0078    0.0034    0.0041    0.0073   -0.0002   -0.0002
        # ...
        # ...
        #         58 H18   1s      0.2678
        #         59 H19   1s     -0.2473
        #         60 H20   1s      0.1835
        #  --
        if '++    Molecular orbitals:' in line:

            self.skip_lines(inputfile, ['d', 'b'])
            line = next(inputfile)

            # We don't currently support parsing natural orbitals or active space orbitals.
            if 'Natural orbitals' not in line and "Pseudonatural" not in line:
                self.skip_line(inputfile, 'b')

                # Symmetry is not currently supported, so this line can have one form.
                while 'Molecular orbitals for symmetry species 1: a' not in line.strip():
                    line = next(inputfile)

                # Symmetry is not currently supported, so this line can have one form.
                if line.strip() != 'Molecular orbitals for symmetry species 1: a':
                    return
                
                line = next(inputfile)
                moenergies = []
                homos = 0
                mocoeffs = []
                while line[:2] != '--':
                    line = next(inputfile)
                    if line.strip().startswith('Orbital'):
                        orbital_index = line.split()[1:]
                        for i in orbital_index:
                            mocoeffs.append([])

                    if 'Energy' in line:
                        energies = [utils.convertor(float(x), 'hartree', 'eV') for x in line.split()[1:]]
                        moenergies.extend(energies)

                    if 'Occ. No.' in line:
                        for i in line.split()[2:]:
                            if float(i) != 0:
                                homos += 1

                    aonames = []
                    tokens = line.split()
                    if tokens and tokens[0] == '1':
                        while tokens and tokens[0] != '--':
                            aonames.append(f"{tokens[1]}_{tokens[2]}")
                            info = tokens[3:]
                            j = 0
                            for i in orbital_index:
                                mocoeffs[int(i)-1].append(float(info[j]))
                                j += 1
                            line = next(inputfile)
                            tokens = line.split()
                        self.set_attribute('aonames', aonames)

                if len(moenergies) != self.nmo:
                    moenergies.extend([numpy.nan for x in range(self.nmo - len(moenergies))])

                self.append_attribute('moenergies', moenergies)

                if not hasattr(self, 'homos'):
                    self.homos = []
                self.homos.extend([homos-1])

                while len(mocoeffs) < self.nmo:
                    nan_array = [numpy.nan for i in range(self.nbasis)]
                    mocoeffs.append(nan_array)

                self.append_attribute('mocoeffs', mocoeffs)

        ## Parsing MP energy from the &MBPT2 module.
        #  Conventional algorithm used...
        #
        #         SCF energy                           =      -74.9644564043 a.u.
        #         Second-order correlation energy      =       -0.0364237923 a.u.
        #
        #         Total energy                         =      -75.0008801966 a.u.
        #         Reference weight ( Cref**2 )         =        0.98652
        #
        #  ::    Total MBPT2 energy                              -75.0008801966
        #
        #
        #         Zeroth-order energy (E0)             =      -36.8202538520 a.u.
        #
        #         Shanks-type energy S1(E)             =      -75.0009150108 a.u.
        if 'Total MBPT2 energy' in line:
            mpenergies = []
            mpenergies.append(utils.convertor(utils.float(line.split()[4]), 'hartree', 'eV'))
            if not hasattr(self, 'mpenergies'):
                self.mpenergies = []
            self.mpenergies.append(mpenergies)

        # Parsing data ccenergies from &CCSDT module.
        #  --- Start Module: ccsdt at Thu Jul 26 14:03:23 2018 ---
        #
        #  ()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()()
        #
        #                                                 &CCSDT
        # ...
        # ...
        #          14          -75.01515915      -0.05070274      -0.00000029
        #          15          -75.01515929      -0.05070289      -0.00000014
        #          16          -75.01515936      -0.05070296      -0.00000007
        #       Convergence after                    17  Iterations
        #
        #
        #      Total energy (diff) :     -75.01515936      -0.00000007
        #      Correlation energy  :        -0.0507029554992
        if 'Start Module: ccsdt' in line:
            self.skip_lines(inputfile, ['b', '()', 'b'])
            line = next(inputfile)
            if '&CCSDT' in line:
                while not line.strip().startswith('Total energy (diff)'):
                    line = next(inputfile)

                ccenergies = utils.convertor(utils.float(line.split()[4]), 'hartree', 'eV')
                if not hasattr(self, 'ccenergies'):
                    self.ccenergies= []
                self.ccenergies.append(ccenergies)

        #  ++    Primitive basis info:
        #        ---------------------
        #
        #
        #                      *****************************************************
        #                      ******** Primitive Basis Functions (Valence) ********
        #                      *****************************************************
        #
        #
        #   Basis set:C.AUG-CC-PVQZ.........                                                          
        #
        #                    Type         
        #                     s
        #             No.      Exponent    Contraction Coefficients
        #             1  0.339800000D+05   0.000091  -0.000019   0.000000   0.000000   0.000000   0.000000
        #             2  0.508900000D+04   0.000704  -0.000151   0.000000   0.000000   0.000000   0.000000
        # ...
        # ...
        #             29  0.424000000D+00   0.000000   1.000000
        #
        #   Number of primitives                                   93
        #   Number of basis functions                              80
        #
        #  --
        if line.startswith('++    Primitive basis info:'):
            self.skip_lines(inputfile, ['d', 'b', 'b', 's', 'header', 's', 'b'])
            line = next(inputfile)
            gbasis_array = []
            while '--' not in line and '****' not in line:
                if 'Basis set:' in line:
                    basis_element_patterns = re.findall(r'Basis set:([A-Za-z]{1,2})\.', line)
                    assert len(basis_element_patterns) == 1
                    basis_element = basis_element_patterns[0].title()
                    gbasis_array.append((basis_element, []))

                if 'Type' in line:
                    line = next(inputfile)
                    shell_type = line.split()[0].upper()

                    self.skip_line(inputfile, 'headers')
                    line = next(inputfile)

                    exponents = []
                    coefficients = []
                    func_array = []
                    while line.split():
                        exponents.append(utils.float(line.split()[1]))
                        coefficients.append([utils.float(i) for i in line.split()[2:]])
                        line = next(inputfile)

                    for i in range(len(coefficients[0])):
                        func_tuple = (shell_type, [])
                        for iexp, exp in enumerate(exponents):
                            coeff = coefficients[iexp][i]
                            if coeff != 0:
                                func_tuple[1].append((exp, coeff))
                        gbasis_array[-1][1].append(func_tuple)

                line = next(inputfile)

            atomsymbols = [self.table.element[atomno] for atomno in self.atomnos]
            self.gbasis = [[] for i in range(self.natom)]
            for element, gbasis in gbasis_array:
                mask = [element == possible_element for possible_element in atomsymbols]
                indices = [i for (i, x) in enumerate(mask) if x]
                for index in indices:
                    self.gbasis[index] = gbasis

        #  ++    Basis set information:
        #        ----------------------
        # ...
        #        Basis set label: MO.ECP.HAY-WADT.5S6P4D.3S3P2D.14E-LANL2DZ.....
        #
        #        Electronic valence basis set:
        #        ------------------
        #        Associated Effective Charge  14.000000 au
        #        Associated Actual Charge     42.000000 au
        #        Nuclear Model: Point charge
        # ...
        #
        #        Effective Core Potential specification:
        #        =======================================
        #
        #         Label   Cartesian Coordinates / Bohr
        #
        #   MO                 0.0006141610       -0.0006141610        0.0979067106
        #  --
        if '++    Basis set information:' in line:
            self.core_array = []
            basis_element = None
            ncore = 0

            while line[:2] != '--':
                if 'Basis set label' in line:
                    try:
                        basis_element = line.split()[3].split('.')[0]
                        basis_element = basis_element[0] + basis_element[1:].lower()
                    except:
                        self.logger.warning('Basis set label is missing!')
                        basis_element = ''
                if 'valence basis set:' in line.lower():
                    self.skip_line(inputfile, 'd')
                    line = next(inputfile)
                    if 'Associated Effective Charge' in line:
                        effective_charge = float(line.split()[3])
                        actual_charge = float(next(inputfile).split()[3])
                        element = self.table.element[int(actual_charge)]
                        ncore = int(actual_charge - effective_charge)
                        if basis_element:
                            assert basis_element == element
                        else:
                            basis_element = element

                if basis_element and ncore:
                    self.core_array.append((basis_element, ncore))
                    basis_element = ''
                    ncore = 0

                line = next(inputfile)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

"""A reader for XYZ (Cartesian coordinate) files."""

from cclib.io import filereader
from cclib.parser.data import ccData
from cclib.parser.utils import PeriodicTable


class XYZ(filereader.Reader):
    """A reader for XYZ (Cartesian coordinate) files."""

    def __init__(self, source, *args, **kwargs) -> None:
        super().__init__(source, *args, **kwargs)

        self.pt = PeriodicTable()

    def parse(self) -> ccData:
        super().parse()

        self.generate_repr()

        return self.data

    def generate_repr(self) -> None:
        """Convert the raw contents of the source into the internal representation."""

        assert hasattr(self, 'filecontents')

        it = iter(self.filecontents.splitlines())

        # Ordering of lines:
        # 1. number of atoms
        # 2. comment line
        # 3. line of at least 4 columns: 1 is atomic symbol (str), 2-4 are atomic coordinates (float)
        #    repeat for numver of atoms
        # (4. optional blank line)
        # repeat for multiple sets of coordinates

        all_atomcoords = []
        comments = []

        while True:

            try:
                line = next(it)
                if line.strip() == '':
                    line = next(it)
                tokens = line.split()
                assert len(tokens) >= 1
                natom = int(tokens[0])

                comments.append(next(it))

                lines = []
                for _ in range(natom):
                    line = next(it)
                    tokens = line.split()
                    assert len(tokens) >= 4
                    lines.append(tokens)
                assert len(lines) == natom

                atomsyms = [line[0] for line in lines]
                atomnos = [self.pt.number[atomsym] for atomsym in atomsyms]
                atomcoords = [line[1:4] for line in lines]
                # Everything beyond the fourth column is ignored.
                all_atomcoords.append(atomcoords)

            except StopIteration:
                break

        attributes = {
            'natom': natom,
            'atomnos': atomnos,
            'atomcoords': all_atomcoords,
            'metadata': {"comments": comments},
        }

        self.data = ccData(attributes)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

"""Generic file writer and related tools"""

import logging
from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import List, Optional, Tuple

import numpy

from cclib.parser.data import ccData
from cclib.parser.utils import PeriodicTable
from cclib.parser.utils import find_package

_has_openbabel = find_package("openbabel")
if _has_openbabel:
    from cclib.bridge import makeopenbabel
    # Open Babel 3.0 and above
    try:
        from openbabel import openbabel as ob
        import openbabel.pybel as pb
    # Open Babel 2.4.x and below
    except:
        import openbabel as ob
        import pybel as pb


class MissingAttributeError(Exception):
    pass


class Writer(ABC):
    """Abstract class for writer objects."""

    required_attrs: Tuple[str] = ()

    def __init__(self, ccdata: ccData, jobfilename: Optional[str] = None, indices=None, terse: bool = False,
                 *args, **kwargs) -> None:
        """Initialize the Writer object.

        This should be called by a subclass in its own __init__ method.

        Inputs:
          ccdata - An instance of ccData, parsed from a logfile.
          jobfilename - The filename of the parsed logfile.
          indices - One or more indices for extracting specific geometries/etc. (zero-based)
          terse - Whether to print the terse version of the output file - currently limited to cjson/json formats
        """

        self.ccdata = ccdata
        self.jobfilename = jobfilename
        self.indices = indices
        self.terse = terse
        self.ghost = kwargs.get("ghost")
        self.naturalorbitals = kwargs.get("naturalorbitals")

        self.pt = PeriodicTable()

        self._check_required_attributes()

        # Open Babel isn't necessarily present.
        if _has_openbabel:
            # Generate the Open Babel/Pybel representation of the molecule.
            # Used for calculating SMILES/InChI, formula, MW, etc.
            self.obmol, self.pbmol = self._make_openbabel_from_ccdata()
            self.bond_connectivities = self._make_bond_connectivity_from_openbabel(self.obmol)

        self._fix_indices()

    @abstractmethod
    def generate_repr(self) -> str:
        """Generate the written representation of the logfile data."""

    def _calculate_total_dipole_moment(self) -> Optional[numpy.ndarray]:
        """Calculate the total dipole moment."""

        # ccdata.moments may exist, but only contain center-of-mass coordinates
        if len(getattr(self.ccdata, 'moments', [])) > 1:
            return numpy.linalg.norm(self.ccdata.moments[1])
        return None

    def _check_required_attributes(self) -> None:
        """Check if required attributes are present in ccdata."""
        missing = [x for x in self.required_attrs
                   if not hasattr(self.ccdata, x)]
        if missing:
            raise MissingAttributeError(
                f"Could not parse required attributes to write file: {missing}")

    def _make_openbabel_from_ccdata(self):
        """Create Open Babel and Pybel molecules from ccData."""
        logger = logging.getLogger("cclib")
        if not hasattr(self.ccdata, "atomcoords"):
            logger.warning("ccdata object does not have atomic coordinates, setting to None")
            _atomcoords = None
        else:
            _atomcoords = self.ccdata.atomcoords
        if not hasattr(self.ccdata, "atomnos"):
            logger.warning("ccdata object does not have atomic numbers, setting to None")
            _atomnos = None
        else:
            _atomnos = self.ccdata.atomnos
        if not hasattr(self.ccdata, 'charge'):
            logger.warning("ccdata object does not have charge, setting to 0")
            _charge = 0
        else:
            _charge = self.ccdata.charge
        if not hasattr(self.ccdata, 'mult'):
            logger.warning("ccdata object does not have spin multiplicity, setting to 1")
            _mult = 1
        else:
            _mult = self.ccdata.mult
        obmol = makeopenbabel(atomcoords=_atomcoords,
                              atomnos=_atomnos,
                              charge=_charge,
                              mult=_mult)
        if self.jobfilename is not None:
            obmol.SetTitle(self.jobfilename)
        return (obmol, pb.Molecule(obmol))
    def _make_bond_connectivity_from_openbabel(self, obmol) -> List[Tuple[int, int, int]]:
        """Based upon the Open Babel/Pybel molecule, create a list of tuples
        to represent bonding information, where the three integers are
        the index of the starting atom, the index of the ending atom,
        and the bond order.
        """
        bond_connectivities = []
        for obbond in ob.OBMolBondIter(obmol):
            bond_connectivities.append((obbond.GetBeginAtom().GetIndex(),
                                        obbond.GetEndAtom().GetIndex(),
                                        obbond.GetBondOrder()))
        return bond_connectivities

    def _fix_indices(self) -> None:
        """Clean up the index container type and remove zero-based indices to
        prevent duplicate structures and incorrect ordering when
        indices are later sorted.
        """
        if not self.indices:
            self.indices = set()
        elif not isinstance(self.indices, Iterable):
            self.indices = set([self.indices])
        # This is the most likely place to get the number of
        # geometries from.
        if hasattr(self.ccdata, 'atomcoords'):
            lencoords = len(self.ccdata.atomcoords)
            indices = set()
            for i in self.indices:
                if i < 0:
                    i += lencoords
                indices.add(i)
            self.indices = indices
        return


del find_package



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

"""Calculate properties for electrons."""

import logging

import numpy

from cclib.method.calculationmethod import Method


class Electrons(Method):
    """A container for methods pertaining to electrons."""

    def __init__(self, data, progress=None, loglevel=logging.INFO, logname="Log"):
        super().__init__(data, progress, loglevel, logname)
        self.required_attrs = ('atomnos','charge','coreelectrons','homos')

    def __str__(self):
        """Returns a string representation of the object."""
        return "Electrons"

    def __repr__(self):
        """Returns a representation of the object."""
        return "Electrons"

    def alpha(self):
        """Number of alpha electrons"""
        return self.data.homos[0] + 1

    def beta(self):
        """Number of beta electrons"""
        return self.data.homos[-1] + 1
        
    def count(self, core=False):
        """Returns the electron count in system.

        Normally returns electrons used in calculation, but will include
        core electrons in pseudopotentials if core is True.
        """
        nelectrons = sum(self.data.atomnos) - self.data.charge
        if core:
            nelectrons += sum(self.data.coreelectrons)
        return nelectrons



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

"""Generic file reader and related tools"""

from abc import ABC, abstractmethod
import typing

from cclib.parser.logfilewrapper import FileWrapper


class Reader(ABC):
    """Abstract class for reader objects."""

    def __init__(self,
        source: typing.Union[str, typing.IO, FileWrapper, typing.List[typing.Union[str, typing.IO]]],
        *args,
        **kwargs)-> None:
        """Initialize the Reader object.

        This should be called by a subclass in its own __init__ method.

        Inputs:
          source - A single filename, stream [TODO], or list of filenames/streams [TODO].
        """
        if not isinstance(source, FileWrapper):
            source = FileWrapper(source)
        
        self.inputfile = source

    def parse(self) -> None:
        """Read the raw contents of the source into the Reader."""
        self.filecontents = self.inputfile.read()

    @abstractmethod
    def generate_repr(self) -> None:
        """Convert the raw contents of the source into the internal representation."""



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

# TOOD: This file belongs in cclib.io, but circular dependency issues mean it can't go there just now.

import bz2
import gzip
import zipfile
import pathlib
from tempfile import NamedTemporaryFile
from urllib.request import urlopen
from urllib.error import URLError
import collections
import typing
import re
import io
import logging
import codecs


# Regular expression for validating URLs
URL_PATTERN = re.compile(

    r'^(?:http|ftp)s?://'  # http:// or https://
    r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
    r'localhost|'  # localhost...
    r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
    r'(?::\d+)?'  # optional port
    r'(?:/?|[/?]\S+)$', re.IGNORECASE

)


def logerror(error):
    """
    Log a unicode decode/encode error to the logger and return a replacement character.
    """
    logging.getLogger("cclib").warning(str(error))
    
    # Return type is a tuple.
    # First item is a replacement character. Second is the position to continue from.
    return (u'', error.start +1)
    
codecs.register_error('logerror', logerror)


class FileWrapper:
    """Wrap any supported input file type."""

    def __init__(self, *sources) -> None:
        # The total size of all our files.
        self.size = 0
        # Current 'byte' position in all our files.
        self.pos = 0
        # The current file position.
        self.file_pointer = 0
        
        self.filenames = []
        self.files = []
        self.sizes = []
        
        # First, check if we were given an unpacked list for backwards compatibility.
        expanded_sources = []
        for source in sources:
            if isinstance(source, list):
                expanded_sources.extend(source)
            
            else:
                expanded_sources.append(source)
        
        # Each source can be a lot of different things, go through and process them now.
        for source in expanded_sources:
            # 'open' the file. If the file is compressed, this function will uncompress it.
            # Likewise, appropriate decoding and error handling will be applied.
            #
            # If no file has been opened yet (source is a string-like), open it.
            filename, fileobject = self.open_log_file(source)
            
            # Move to the end of the file to determine how big it is.
            fileobject.seek(0, 2)
            size = fileobject.tell()
            fileobject.seek(0, 0)
            
            self.files.append(fileobject)
            # open_log_file returns a pathlib.Path object, cast to str for compatibility.
            self.filenames.append(str(filename))
            self.sizes.append(size)
        
        # Total number of bytes in all our files.
        self.size = sum(self.sizes)
        
        # A short buffer of previously read lines.
        # This permits primitive 'look-behind' functionality in some parsers (see Turbomole).
        # Init with 10 empty strings (empty lines).
        self.last_lines = collections.deque([""] * 10, 10)
        
    @property
    def file_name(self) -> str:
        return ", ".join(self.filenames)
    
    def sort(self, order: list) -> None:
        """
        Sort the individual files that make up this FileWrapper.
        
        order is an ordered list of filenames.
        """
        # Reset all our streams.
        self.reset()
        
        filenames = []
        files = []
        sizes = []
        
        # Reorder.
        for filename in order:
            index = self.filenames.index(filename)
            filenames.append(filename)
            files.append(self.files[index])
            sizes.append(self.sizes[index])
            
        self.filenames = filenames
        self.files = files
        self.sizes = sizes
    
    @classmethod
    def open_log_file(
            self,
            source,
            mode: str = "r",
            encoding: str = "utf-8",
            errors: str = "logerror",
        ) -> typing.Tuple[str, typing.IO]:
        """
        Open a possibly compressed file, returning both the filename of the file and an open file object.
        """
        # First, work out what source is (could be a filename, a URL, an open file etc).
        if isinstance(source, str) and URL_PATTERN.match(source):
            # This file is a URL.
            try:
                # Cache the file to a temp location.
                response = urlopen(source)
                fileobject = NamedTemporaryFile(delete = True)
                fileobject.write(response.read())
                fileobject.seek(0,0)
                
                fileobject = io.TextIOWrapper(fileobject, encoding = encoding, errors = errors)
                filename = source
                
            except (ValueError, URLError) as error:
                # Maybe no need to raise a different exception?
                raise ValueError(
                    "Encountered an error processing the URL '{}'".format(source)
                ) from error
                
        elif hasattr(source, "read") or hasattr(source, "readline"):
            # This file is a file.
            # If this file supports seek, we don't need to do anything.
            # If not, we'll cache it to file.
            if not hasattr(source, "seek") or \
            (hasattr(source, "seekable") and not source.seekable()):
                fileobject = NamedTemporaryFile(delete = True)
                fileobject.write(source.read())
                fileobject.seek(0,0)
                
                fileobject = io.TextIOWrapper(fileobject, encoding = encoding, errors = errors)
             
            else:
                fileobject = source
            filename = getattr(source, "name", f"stream {str(type(source))}")
            
        else:
            # This file is something else, assume we can open() it.
            filename = source
            fileobject = None
        
        filename = pathlib.Path(filename)
        extension = filename.suffix
    
        if extension == ".gz":
            fileobject = io.TextIOWrapper(gzip.GzipFile(filename, mode, fileobj = fileobject), encoding = encoding, errors = errors)
    
        elif extension == ".zip":
            fileobject = zipfile.ZipFile(fileobject if fileobject else filename, mode)
            # TODO: Need to check that we're not leaving any open file objects here...
            # TODO: We should be able to handle multiple files...
            assert len(fileobject.namelist()) == 1, "ERROR: Zip file contains more than 1 file"
            
            fileobject = io.TextIOWrapper(
                fileobject.open(fileobject.namelist()[0]),
                encoding = encoding, errors = errors
            )
    
        elif extension in ['.bz', '.bz2']:
            # Module 'bz2' is not always importable.
            assert bz2 is not None, "ERROR: module bz2 cannot be imported"
            fileobject = io.TextIOWrapper(bz2.BZ2File(fileobject if fileobject else filename, mode), encoding = encoding, errors = errors)
    
        elif fileobject is not None:
            # Assuming that object is text file encoded in utf-8
            # If the file/stream has already been opened, we have no ability to handle decoding errors.
            pass
            
        else:
            # Normal text file.
            
            fileobject = open(filename, mode, encoding = encoding, errors = errors)
        
        return filename, fileobject

    def next(self) -> str:
        """
        Get the next line from this log file.
        """
        try:
            try:
                line = next(self.files[self.file_pointer])
                self.last_lines.append(line)
                self.pos += len(line)
                return line
            
            except StopIteration:
                self.file_pointer += 1
                return self.next()
            
        except IndexError:
            raise StopIteration()
    
    @property
    def last_line(self) -> str:
        """
        Return the last line read by this parser.
        """
        return self.last_lines[-1]

    def __next__(self):
        return self.next()

    def __iter__(self):
        return self

    # TODO: support size parameter.
    def readline(self) -> str:
        """
        Read one line from this file.
        """
        return next(self)
    
    # TODO: support size parameter.
    def read(self) -> str:
        """
        Read everything from this file.
        
        Be aware that this function will load the entire file into a single string.
        """
        return "".join(list(self))

    def close(self) -> None:
        """
        Close all open files.
        """
        for file in self.files:
            file.close()
            
    def __del__(self) -> None:
        """
        Make sure to close any open files when we go out of scope.
        
        Note that there is no guarantee when or if this function will get called;
        user's should ensure to close their own files once they are finished with them.
        """
        self.close()

    def seek(self, offset: int, whence: int = 0) -> None:
        if offset != 0 or whence not in (0,2):
            raise NotImplementedError("FileWrapper only supports seeking to start or end")
        
        if whence == 0:
            self.reset()
        
        elif whence == 2:
            self.finish()

#     def seek_from_current(self, offset):
#         """
#         Seek forwards or backwards based on the current position.
#         """
#         remainder = offset
#         while remainder != 0:
#             if remainder > 0:
#                 # Seek forwards please.
#                 file_size = self.sizes[self.file_pointer]
            
    
    def reset(self):
        # Equivalent to seeking to 0 for all our files.
        for file in self.files:
            file.seek(0,0)
            
        self.file_pointer = 0
        self.pos = 0
        
    def finish(self):
        # Equivalent to seeking to 2 for all our files.
        for file in self.files:
            file.seek(0,2)
            
        self.file_pointer = len(self.files) -1
        self.pos = self.size



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

"""Parser for Formatted Checkpoint files"""


import numpy

from cclib.parser import logfileparser
from cclib.parser import utils

SHELL_ORBITALS = {
    0: ['S'],
    1: ['PX', 'PY', 'PZ'],
    -1: ['S', 'PX', 'PY', 'PZ'],
    2: ['D1', 'D2', 'D3', 'D4', 'D5', 'D6'],
    -2: ['D1', 'D2', 'D3', 'D4', 'D5'],
    3:  ['F1', 'F2', 'F3', 'F4', 'F5', 'F6', 'F7', 'F8', 'F9', 'F10'],
    -3: ['F1', 'F2', 'F3', 'F4', 'F5', 'F6', 'F7'],
    4: ['G1', 'G2', 'G3', 'G4', 'G5', 'G6', 'G7', 'G8', 'G9', 'G10', 'G11', 'G12','G13'],
    -4: ['G1', 'G2', 'G3', 'G4', 'G5', 'G6', 'G7', 'G8', 'G9']

}

SHELL_START = {
    0: 1,
    1: 2,
    -1: 2,
    2: 3,
    -2: 3,
    3: 4,
    -3: 4
}


def _shell_to_orbitals(type, offset):
    """Convert a Fchk shell type and offset to a list of string representations.

    For example, shell type = -2 corresponds to d orbitals (spherical basis) with
    an offset = 1 would correspond to the 4d orbitals, so this function returns
    `['4D1', '4D2', '4D3', '4D4', '4D5']`.
    """

    return [f"{SHELL_START[type] + offset}{x}" for x in SHELL_ORBITALS[type]]


class FChk(logfileparser.Logfile):
    """A Formatted checkpoint file, which contains molecular and wavefunction information.

    These files are produced by Gaussian and Q-Chem.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(logname="FChk", *args, **kwargs)
        self.start = True

    def __str__(self):
        """Return a string representation of the object."""
        return f"Formatted checkpoint file {self.filename}"

    def __repr__(self):
        """Return a representation of the object."""
        return f'FCHK("{self.filename}")'

    def normalisesym(self, symlabel):
        """Just return label"""
        return symlabel

    def extract(self, inputfile, line):

        # just opened file, skip first line to get basis
        if self.start:
            method = next(inputfile)
            self.metadata['basis_set'] = method.split()[-1]
            self.start = False

        if line[0:6] == 'Charge':
            self.set_attribute('charge', int(line.split()[-1]))

        if line[0:12] == 'Multiplicity':
            self.set_attribute('mult', int(line.split()[-1]))

        if line[0:14] == 'Atomic numbers':
            self.natom = int(line.split()[-1])
            atomnos = self._parse_block(inputfile, self.natom, int, 'Basic Information')
            self.set_attribute('atomnos', atomnos)

        if line[0:19] == 'Number of electrons':
            alpha = next(inputfile)
            alpha_homo = int(alpha.split()[-1]) - 1

            beta = next(inputfile)
            beta_homo = int(beta.split()[-1]) - 1

            self.set_attribute('homos', [alpha_homo, beta_homo])

        if line[0:29] == 'Current cartesian coordinates':
            count = int(line.split()[-1])
            assert count % 3 == 0

            coords = numpy.array(self._parse_block(inputfile, count, float, 'Coordinates'))
            coords.shape = (1, int(count / 3), 3)
            self.set_attribute('atomcoords', utils.convertor(coords, 'bohr', 'Angstrom'))

        if line[0:10] == "SCF Energy":
            self.set_attribute(
                "scfenergies",
                [utils.convertor(float(line.split()[-1]), "hartree", "eV")]
            )

        if line[0:25] == 'Number of basis functions':
            self.set_attribute('nbasis', int(line.split()[-1]))

        if line[0:14] == 'Overlap Matrix':
            count = int(line.split()[-1])
            self.set_attribute(
                "aooverlaps",
                utils.block_to_matrix(
                    numpy.asarray(
                        self._parse_block(inputfile, count, float, "Overlap Matrix")
                    )
                )
            )

        if line[0:31] == 'Number of independent functions':
            self.set_attribute('nmo', int(line.split()[-1]))

        if line[0:21] == 'Alpha MO coefficients':
            count = int(line.split()[-1])
            assert count == self.nbasis * self.nmo

            coeffs = numpy.array(self._parse_block(inputfile, count, float, 'Alpha Coefficients'))
            coeffs.shape = (self.nmo, self.nbasis)
            self.set_attribute('mocoeffs', [coeffs])

        if line[0:22] == 'Alpha Orbital Energies':
            count = int(line.split()[-1])
            assert count == self.nmo

            energies = numpy.array(self._parse_block(inputfile, count, float, 'Alpha MO Energies'))
            self.set_attribute('moenergies', [utils.convertor(energies, 'hartree', 'eV')])

        if line[0:20] == 'Beta MO coefficients':
            count = int(line.split()[-1])
            assert count == self.nbasis * self.nmo

            coeffs = numpy.array(self._parse_block(inputfile, count, float, 'Beta Coefficients'))
            coeffs.shape = (self.nmo, self.nbasis)
            self.append_attribute('mocoeffs', coeffs)

        if line[0:21] == 'Beta Orbital Energies':
            count = int(line.split()[-1])
            assert count == self.nmo

            energies = numpy.array(self._parse_block(inputfile, count, float, 'Alpha MO Energies'))
            self.append_attribute('moenergies', utils.convertor(energies, 'hartree', 'eV'))

        if line[0:11] == 'Shell types':
            self.parse_aonames(line, inputfile)

        if line[0:19] == 'Real atomic weights':
            count = int(line.split()[-1])
            assert count == self.natom

            atommasses = numpy.array(self._parse_block(inputfile, count, float, 'Atomic Masses'))

            self.set_attribute('atommasses', atommasses)

        if line[0:10] == 'SCF Energy':
            self.scfenergy = float(line.split()[-1])

            self.set_attribute('scfenergies', [utils.convertor(self.scfenergy,'hartree','eV')])

        if line[0:16] == "Mulliken Charges":
            count = int(line.split()[-1])
            if not hasattr(self, "atomcharges"):
                self.atomcharges = {}
            self.atomcharges["mulliken"] = self._parse_block(inputfile, count, float, "Mulliken Charges")

        if line[0:18] == 'Cartesian Gradient':
            count = int(line.split()[-1])
            assert count == self.natom*3

            gradient = numpy.array(self._parse_block(inputfile, count, float, 'Gradient'))

            self.set_attribute('grads', gradient)

        if line[0:14] == "Polarizability":
            polarizability = numpy.asarray(self._parse_block(inputfile, 6, float, "Polarizability"))
            self.append_attribute("polarizabilities", utils.block_to_matrix(polarizability))

        if line[0:25] == 'Cartesian Force Constants':
            count = int(line.split()[-1])
            assert count == (3*self.natom*(3*self.natom+1))/2

            hessian = numpy.array(self._parse_block(inputfile, count, float, 'Hessian'))

            self.set_attribute('hessian', utils.block_to_matrix(hessian))

        if line[0:22] == "Number of Normal Modes":
            self.set_attribute("nmode", int(line.split()[-1]))

        if line[0:6] == "Vib-E2":
            count = int(line.split()[-1])
            le2 = self._parse_block(inputfile, count, float, "Vib-E2")
            assert hasattr(self, "nmode")
            nmode = self.nmode
            self.set_attribute("vibfreqs", le2[:nmode])
            le2 = le2[nmode:]
            self.set_attribute("vibrmasses", le2[:nmode])
            le2 = le2[nmode:]
            self.set_attribute("vibfconsts", le2[:nmode])
            le2 = le2[nmode:]
            self.set_attribute("vibirs", le2[:nmode])
            # The rest should be empty unless Raman or ROA were calculated.
            le2 = numpy.asarray(le2[nmode:])
            if numpy.any(le2[:nmode] >= 1.0e-20):
                self.set_attribute("vibramans", le2[:nmode])
                le2 = le2[nmode:]

        if line[0:9] == "Vib-Modes":
            count = int(line.split()[-1])
            vibdisps = numpy.asarray(self._parse_block(inputfile, count, float, "Vib-Modes"))
            # indices from fast to slow are [xyz, atom, mode]
            assert hasattr(self, "nmode")
            assert count == 3 * self.natom * self.nmode
            self.set_attribute("vibdisps", vibdisps.reshape(self.nmode, self.natom, 3))

        if line[0:13] == 'ETran scalars':
            count = int(line.split()[-1])

            etscalars = self._parse_block(inputfile, count, int, 'ET Scalars')

            # Set attribute: self.netroot (number of excited estates)
            self.netroot = etscalars[4]

        if line[0:10] == 'ETran spin':
            count = int(line.split()[-1])

            etspin = self._parse_block(inputfile, count, int, 'ET Spin')

            spin_labels = { 0:'Singlet',
                            2:'Triplet',
                           -1:'Unknown'}
            etsyms = []
            for i in etspin:
                if i in spin_labels:
                    etsyms.append(spin_labels[i])
                else:
                    etsyms.append(spin_labels[-1])

            # The extracted property does not contain the actual irrep label
            # (contrarily to that extracted from the Gaussian log)
            # After this, 'Etran sym' appears (and would need to be parsed), 
            # but at least in Gaussian this contains only zeroes regardless of the irrep.

            self.set_attribute('etsyms', etsyms)

        if line[0:18] == 'ETran state values':
            # This section is organized as follows:
            # ·First the properties of each excited state (up to net):
            # E, {muNx,muNy,muNz,muvelNx,muvelNy,muvelNz,mmagNx,mmagNy,mmagNz,unkX,unkY,unkZ,unkX,unkY,unkZ}_N=1,net
            # ·Then come 48 items (only if Freq is requested)
            # They were all 0.000 in G09, but get an actual value in G16
            # ·Then, the derivates of each property with respect to Cartesian coordiates only for target state (netroot)
            # For each Cartesian coordiate, all derivatives wrt to it are included:
            #  dE/dx1 dmux/dx1 dmuy/dx1 ... unkZ/dx1
            #  dE/dy1 dmux/dy1 dmuy/dy1 ... unkZ/dy1
            #  ...
            #  dE/dzN dmux/dzN dmuy/dzN ... unkZ/dzN
            # The number of items is therefore:
            ### 16*net (no Freq jobs)
            ### 16*net + 48 + 3*self.natom*16 (Freq jobs)
            count = int(line.split()[-1])
            if hasattr(self,'etsyms'):
                net = len(self.etsyms)
            else:
                net = 0 # This forces an AssertionError below
            assert count in [16*net, 16*net+48+3*self.natom*16]

            etvalues = self._parse_block(inputfile, count, float, 'ET Values')

            # ETr energies (1/cm)
            etenergies_au = [ e_es-self.scfenergy for e_es in etvalues[0:net*16:16] ]
            etenergies = [ utils.convertor(etr,'hartree','wavenumber') for etr in etenergies_au ]
            self.set_attribute('etenergies', etenergies)

            # ETr dipoles (length-gauge)
            etdips = []
            for k in range(1,16*net,16):
                etdips.append(etvalues[k:k+3])
            self.set_attribute('etdips',etdips)

            # Osc. Strength from Etr dipoles
            # oscs = 2/3 * Etr(au) * dip²(au)
            etoscs = [ 2.0/3.0*e*numpy.linalg.norm(numpy.array(dip))**2 for e,dip in zip(etenergies_au,etdips) ]
            self.set_attribute('etoscs', etoscs)

            # ETr dipoles (velocity-gauge)
            etveldips = []
            for k in range(4,16*net,16):
                etveldips.append(etvalues[k:k+3])
            self.set_attribute('etveldips',etveldips)

            # ETr magnetic dipoles
            etmagdips = []
            for k in range(7,16*net,16):
                etmagdips.append(etvalues[k:k+3])
            self.set_attribute('etmagdips',etmagdips)

        if line[0:19] == "Excitation Energies":
            count = int(line.split()[-1])
            etenergies = self._parse_block(inputfile, count, float, "Excitation energies")
            etenergies = [utils.convertor(e, "hartree", "wavenumber") for e in etenergies]
            self.set_attribute("etenergies", etenergies)

        if line[0:20] == "Oscillator Strengths":
            count = int(line.split()[-1])
            etoscs = self._parse_block(inputfile, count, float, "Oscillator Strengths")
            self.set_attribute("etoscs", etoscs)

    def parse_aonames(self, line, inputfile):
        # e.g.: Shell types                                I   N=          28
        count = int(line.split()[-1])
        shell_types = self._parse_block(inputfile, count, int, 'Atomic Orbital Names')

        # e.g.: Number of primitives per shell             I   N=          28
        next(inputfile)
        self._parse_block(inputfile, count, int, 'Atomic Orbital Names')

        # e.g. Shell to atom map                          I   N=          28
        next(inputfile)
        shell_map = self._parse_block(inputfile, count, int, 'Atomic Orbital Names')

        elements = (self.table.element[x] for x in self.atomnos)
        atom_labels = [f"{y}{x}" for x, y in enumerate(elements, 1)]

        # get orbitals for first atom and start aonames and atombasis lists
        atom = shell_map[0] - 1
        shell_offset = 0
        orbitals = _shell_to_orbitals(shell_types[0], shell_offset)
        aonames = [f"{atom_labels[atom]}_{x}" for x in orbitals]
        atombasis = [list(range(len(orbitals)))]

        # get rest
        for i in range(1, len(shell_types)):
            _type = shell_types[i]
            atom = shell_map[i] - 1
            shell_offset += 1
            basis_offset = atombasis[-1][-1] + 1 # atombasis is increasing numbers, so just grab last

            # if we've move to next atom, need to update offset of shells (e.g. start at 1S)
            # and start new list for atom basis
            if atom != shell_map[i - 1] - 1:
                shell_offset = 0
                atombasis.append([])

            # determine if we've changed shell type (e.g. from S to P)
            if _type != shell_types[i - 1]:
                shell_offset = 0

            orbitals = _shell_to_orbitals(_type, shell_offset)
            aonames.extend([f"{atom_labels[atom]}_{x}" for x in orbitals])
            atombasis[-1].extend(list(range(basis_offset, basis_offset + len(orbitals))))

        assert (
            len(aonames) == self.nbasis
        ), f"Length of aonames != nbasis: {len(aonames)} != {self.nbasis}"
        self.set_attribute("aonames", aonames)

        assert (
            len(atombasis) == self.natom
        ), f"Length of atombasis != natom: {len(atombasis)} != {self.natom}"
        self.set_attribute("atombasis", atombasis)

    def after_parsing(self):
        """Correct data or do parser-specific validation after parsing is finished."""

        # If restricted calculation, need to remove beta homo
        if len(self.moenergies) == len(self.homos) - 1:
            self.homos.pop()

    def _parse_block(self, inputfile, count, type, msg):
        atomnos = []
        while len(atomnos) < count :
            self.updateprogress(inputfile, msg, self.fupdate)
            line = next(inputfile)
            atomnos.extend([ type(x) for x in line.split()])
        return atomnos



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

"""Test coupled cluster logfiles"""

import os
import unittest

import numpy
import pytest

__filedir__ = os.path.realpath(os.path.dirname(__file__))


class GenericCCTest(unittest.TestCase):
    """Generic coupled cluster unittest"""

    rel_thresh = 1.0e-4

    def testsizeandshape(self):
        """Are the dimensions of ccenergies correct?"""
        assert self.data.ccenergies.shape == \
                         (len(self.data.scfenergies),)

    def testsign(self):
        """Are the coupled cluster corrections negative?"""
        corrections = self.data.ccenergies - self.data.scfenergies
        assert numpy.alltrue(corrections < 0.0)


class GenericCC2Test(GenericCCTest):

    # Turbomole 7.4
    corr_energy = -1.1508051574141973

    def testenergycc2(self):
        """Is the CC2 correlation energy within the target?"""
        e_scf = self.data.scfenergies[0]
        e_cc = self.data.ccenergies[0]
        e_corr = e_cc - e_scf
        assert pytest.approx(e_corr, rel=self.rel_thresh) == self.corr_energy


class GenericCCDTest(GenericCCTest):

    # Q-Chem 5.4
    corr_energy = -1.4435403900740766

    def testenergyccd(self):
        """Is the CCD correlation energy within the target?"""
        e_scf = self.data.scfenergies[0]
        e_cc = self.data.ccenergies[0]
        e_corr = e_cc - e_scf
        assert pytest.approx(e_corr, rel=self.rel_thresh) == self.corr_energy


class GenericCCSDTest(GenericCCTest):

    # Q-Chem 5.4
    corr_energy = -1.4518567335733223

    def testenergyccsd(self):
        """Is the CCSD correlation energy within the target?"""
        e_scf = self.data.scfenergies[0]
        e_cc = self.data.ccenergies[0]
        e_corr = e_cc - e_scf
        assert pytest.approx(e_corr, rel=self.rel_thresh) == self.corr_energy


class GenericCCSDPTTest(GenericCCTest):

    # Q-Chem 5.4
    corr_energy = -1.4539460237174353

    def testenergyccsdpt(self):
        """Is the CCSD(T) correlation energy within the target?"""
        e_scf = self.data.scfenergies[0]
        e_cc = self.data.ccenergies[0]
        e_corr = e_cc - e_scf
        assert pytest.approx(e_corr, rel=self.rel_thresh) == self.corr_energy


class DALTONCCSDPTTest(GenericCCSDPTTest):

    # DALTON 2015
    corr_energy = -0.39526637011522325


class GamessCCDTest(GenericCCDTest):

    # GAMESS-US 2018
    corr_energy = -5.633089378137811


class GamessCCSDTest(GenericCCSDTest):

    # GAMESS-US 2018
    corr_energy = -5.655488429741581


class GamessCCSDPTTest(GenericCCSDPTTest):

    # GAMESS-US 2018
    corr_energy = -5.734634247949089


class MolcasCCSDTest(GenericCCSDTest):

    # OpenMolcas 18.0
    corr_energy = -1.3796976452617855


class MolproCCDTest(GenericCCDTest):

    # Molpro 2012
    corr_energy = -5.76891848850255


class MolproCCSDTest(GenericCCSDTest):

    # Molpro 2012
    corr_energy = -5.790734939563208


class MolproCCSDPTTest(GenericCCSDPTTest):

    # Molpro 2012
    corr_energy = -5.876248590504929


class NWChemCCSDPTTest(GenericCCSDPTTest):

    # NWChem 7.0
    corr_energy = -1.4559748390679488


class OrcaCCSDTest(GenericCCSDTest):

    # ORCA 5.0
    corr_energy = -1.3582212909295777


class OrcaCCSDPTTest(GenericCCSDPTTest):

    # ORCA 5.0
    corr_energy = -1.360085161959887


class Psi4CCSDTest(GenericCCSDTest):

    # Psi4 1.3.1
    corr_energy = -1.4538882732524598


class Psi4CCSDPTTest(GenericCCSDPTTest):

    # Psi4 1.3.1
    corr_energy = -1.4559754695608262


class TurbomoleCCSDTest(GenericCCSDTest):

    # TODO The two Turbomole CCSD(T) energies aren't similar...
    rel_thresh = 5.0e-3

    # Turbomole 7.4
    corr_energy = -1.3877950772714485


if __name__ == "__main__":

    import sys
    sys.path.insert(1, os.path.join(__filedir__, ".."))

    from test_data import DataSuite
    suite = DataSuite(['CC'])
    suite.testall()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

"""Unit tests for parser utils module."""

import unittest

import cclib


class convertorTest(unittest.TestCase):

    def test_basic(self):
        """Are some basic conversions correct?"""
        convertor = cclib.parser.utils.convertor
        assert round(abs(convertor(1.89, "bohr", "Angstrom") - 1.0), 3) == 0
        assert round(abs(convertor(0.529, "Angstrom", "bohr") - 1.0), 3) == 0
        assert round(abs(convertor(627.5, "kcal/mol", "hartree") - 1.0), 3) == 0

    def test_pairs(self):
        """Do flipped conversions correspond to each other?"""

        convertor = cclib.parser.utils.convertor

        pairs_proportional = (
            ("Angstrom", "bohr"), ("wavenumber", "eV"),
            ("wavenumber", "kcal/mol"), ("eV", "kJ/mol"), ("coulomb", "e"))
        pairs_inverse = (("nm", "wavenumber"),)

        for unit1, unit2 in pairs_proportional:
            conv1 = convertor(1.0, unit1, unit2)
            conv2 = 1.0 / convertor(1.0, unit2, unit1)
            assert round(abs((conv1 - conv2) / conv1), 7) == 0
        for unit1, unit2 in pairs_inverse:
            conv1 = convertor(1.0, unit1, unit2)
            conv2 = convertor(1.0, unit2, unit1)
            assert round(abs((conv1 - conv2) / conv1), 7) == 0


class PeriodicTableTest(unittest.TestCase):

    def setUp(self):
        self.pt = cclib.parser.utils.PeriodicTable()

    def test_elements(self):
        """Does the periodic table give correct elements?"""
        assert self.pt.element[6] == 'C'
        assert self.pt.element[44] == 'Ru'
        assert self.pt.element[0] is None

    def test_numbers(self):
        """Does the periodic table give correct atom numbers?"""
        assert self.pt.number['C'] == 6
        assert self.pt.number['Au'] == 79


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

import unittest
import os, sys
import numpy

from cclib.bridge import cclib2horton
from ..test_data import getdatafile
from cclib.parser.utils import find_package

from numpy.testing import assert_array_almost_equal


class HortonTest(unittest.TestCase):
    """Tests for the Horton bridge in cclib"""

    # Both horton and cclib can read in fchk files. The test routine utilizes this fact
    # and compares the attributes that were directly loaded from each package
    # and the attributes that were converted.

    def setUp(self):
        super().setUp()

        self.data, self.logfile = getdatafile(
            "FChk", "basicGaussian16", ["dvb_un_sp.fchk"]
        )
        datadir = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "..", "..", "data")
        )
        inputfile = os.path.join(
            datadir, "FChk", "basicGaussian16", "dvb_un_sp.fchk"
        )

        if not find_package("iodata"):
            raise ImportError("Must install iodata to run this test")

        from iodata import IOData
        from iodata.orbitals import MolecularOrbitals
        from iodata.api import load_one

        self._hortonver = 3

        self.iodat = load_one(filename=inputfile)

    def test_makehorton(self):
        """ Check that the bridge from cclib to horton works correctly """
        # First use `makehorton` function to generate IOData object converted from cclib ccData
        hortonequiv = cclib2horton.makehorton(self.data)

        # Identify attributes that should be verified
        check = ["spinpol"]  # float or int
        checkArr = ["atcoords", "atnums", "atcorenums"]  # one dimensional arrays

        # IOData in horton 3 uses attr and initializes all possible attributes by default value
        # of None. Thus, hasattr cannot be used here on its own.
        for attr in check:
            if (
                hasattr(self.iodat, attr)
                and hasattr(hortonequiv, attr)
                and getattr(self.iodat, attr) is not None
                and getattr(hortonequiv, attr) is not None
            ):
                assert abs(getattr(self.iodat, attr)-getattr(hortonequiv, attr)) < 1.0e-3

        for attr in checkArr:
            if (
                hasattr(self.iodat, attr)
                and hasattr(hortonequiv, attr)
                and isinstance(getattr(self.iodat, attr), numpy.ndarray)
                and isinstance(getattr(hortonequiv, attr), numpy.ndarray)
            ):
                assert_array_almost_equal(
                    getattr(self.iodat, attr), getattr(hortonequiv, attr), decimal=3
                )

    def test_makecclib(self):
        """ Check that the bridge from horton to cclib works correctly """
        # First use `makecclib` function to generate ccData object converted from horton IOData
        cclibequiv = cclib2horton.makecclib(self.iodat)

        # Identify attributes that should be verified
        check = ["mult"]  # float or int
        checkArr = ["atomcoords", "atomnos", "coreelectrons", "mocoeffs"]  # one dimensional arrays
        checkArrArr = ["polarizability"]  # two dimensional arrays
        checkChg = ["mulliken", "natural"]  # atomcharges attribute is dictionary with these keys

        for attr in check:
            if hasattr(self.data, attr) and hasattr(cclibequiv, attr):
                assert abs(getattr(self.data, attr)-getattr(cclibequiv, attr)) < 1.0e-3

        for attr in checkArr:
            if hasattr(self.data, attr) and hasattr(cclibequiv, attr):
                assert_array_almost_equal(
                    getattr(self.data, attr), getattr(cclibequiv, attr), decimal=3
                )

        for attr in checkArrArr:
            if hasattr(self.data, attr) and hasattr(cclibequiv, attr):
                assert_array_almost_equal(
                    getattr(self.data, attr)[0], getattr(cclibequiv, attr)[0], decimal=3
                )

        if hasattr(self.data, "atomcharges") and hasattr(cclibequiv, "atomcharges"):
            for chg in checkChg:
                if chg in self.data.atomcharges and chg in cclibequiv.atomcharges:
                    assert_array_almost_equal(
                        self.data.atomcharges[chg][0],
                        cclibequiv.atomcharges[chg][0],
                        decimal=3,
                    )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023, the cclib development team
#
# This file is part of cclib (http://cclib.github.io) and is distributed under
# the terms of the BSD 3-Clause License.

"""Bridge for using cclib data in ASE (https://wiki.fysik.dtu.dk/ase/)."""

import numpy as np

from cclib.parser.data import ccData
from cclib.parser.utils import find_package

_found_ase = find_package("ase")
if _found_ase:
    from ase import Atoms, units
    from ase.io.trajectory import Trajectory
    from ase.calculators.calculator import PropertyNotImplementedError


def _check_ase(found_ase):
    if not found_ase:
        raise ImportError("You must install `ase` to use this function")


def makease(
    atomcoords, atomnos, atomcharges=None, atomspins=None, atommasses=None
):
    """Create an ASE Atoms object from cclib attributes.

    ASE requires atomic partial charges and atomic spin densities rather than
    molecular charge and multiplicity, so we follow how other interfaces have
    done (e.g., MOPAC, Gaussian and XTB) and require atomcharges and atomspins,
    or leave undefined.

    Inputs:
        atomcoords - two dimensional array-like with atomic coordinates.
        atomnos - one dimensional array-like with atomic numbers.
        atomcharges - one dimensional array-like with atomic charges.
        atomspins - one dimensional array-like with atomic spin densities.
        atommasses - one dimensional array-like with atomic masses.
    """
    _check_ase(_found_ase)
    return Atoms(
        positions=atomcoords,
        numbers=atomnos,
        masses=atommasses,
        charges=atomcharges,
        magmoms=atomspins,
    )


def write_trajectory(filename, ccdata, popname="mulliken", index=None):
    """Write an ASE Trajectory object from a ccData object.

    We try to write the following properties: atomcoords, atomnos, atomcharges,
    atomspins, atommasses, scfenergies, grads, moments (dipole only) and
    freeenergy. No charge or mult is written since ASE calculates them from
    atomcharges and atomspins.

    When a program do a final single point calculation at the end of an
    optimization (e.g., ORCA), we'll have unequal number of grads and
    atomcoords. This means the last geometry in the written .traj will lack
    forces.

    Some unit conversions are done here, but things are converted back in
    read_trajectory/makecclib.

    Inputs:
        filename - path to traj file to be written.
        ccdata - an instance of ccData.
        popname - population analysis to use for atomic partial charges and
            atomic spin densities. Molecular charge and multiplicity are
            evaluated from them.
        index - sequence of integers indicating which atomcoords indices should
            be exported. By default, all are exported.
    """
    _check_ase(_found_ase)
    traj = Trajectory(filename, "w")
    for i, atomcoords in enumerate(ccdata.atomcoords):
        if index is not None and i not in index:
            continue

        atomspins = None
        if hasattr(ccdata, "atomspins"):
            atomspins = ccdata.atomspins[popname]
        atoms = makease(
            atomcoords,
            ccdata.atomnos,
            ccdata.atomcharges[popname],
            atomspins,
            ccdata.atommasses,
        )

        properties = {}
        if hasattr(ccdata, "scfenergies"):
            properties.update({"energy": ccdata.scfenergies[i]})
        if hasattr(ccdata, "grads"):
            try:
                properties.update(
                    {"forces": -ccdata.grads[i] * units.Hartree / units.Bohr}
                )
            except IndexError:
                pass

        if i == len(ccdata.atomcoords) - 1:  # last geometry
            if hasattr(ccdata, "moments"):
                properties.update({"dipole": ccdata.moments[1] * units.Bohr})
            if hasattr(ccdata, "free_energy"):
                properties.update(
                    {"free_energy": ccdata.freeenergy * units.Hartree}
                )

        traj.write(atoms, **properties)


def read_trajectory(filename):
    """Read an ASE Trajectory object and return a ccData object.

    The returned object has everything write_trajectory writes, plus natom,
    charge, mult and temperature.

    The following properties are taken from the last frame: atomnos,
    atomcharges, atomspins, atommasses, moments, freeenergy and temperature.
    charge, mult and natom also represent the last frame, since they depend on
    other propertes read from the last frame.

    Bear in mind that ASE calculates temperature from the kinetic energy, so
    anything "static" (which includes anything cclib parses) will have zero
    temperature.

    Inputs:
        filename - path to traj file to be read.
    """
    _check_ase(_found_ase)
    attributes = {"atomcoords": [], "scfenergies": [], "grads": []}

    for atoms in Trajectory(filename, "r"):
        ccdata = makecclib(atoms)
        attributes["atomcoords"].append(ccdata.atomcoords[-1])
        if hasattr(ccdata, "scfenergies"):
            attributes["scfenergies"].append(ccdata.scfenergies[-1])
        if hasattr(ccdata, "grads"):
            attributes["grads"].append(ccdata.grads[-1])

    # ccdata is now last frame
    attributes["atomnos"] = ccdata.atomnos
    attributes["atomcharges"] = ccdata.atomcharges
    attributes["atomspins"] = ccdata.atomspins
    attributes["atommasses"] = ccdata.atommasses

    if hasattr(ccdata, "moments"):
        attributes["moments"] = ccdata.moments
    if hasattr(ccdata, "freeenergy"):
        attributes["freeenergy"] = ccdata.freeenergy

    # remove if empty
    if not attributes["scfenergies"]:
        del attributes["scfenergies"]
    if not attributes["grads"]:
        del attributes["grads"]

    # extra stuff we can't write in write_trajectory
    attributes["temperature"] = ccdata.temperature
    attributes["charge"] = ccdata.charge
    attributes["mult"] = ccdata.mult
    attributes["natom"] = ccdata.natom

    return ccData(attributes)


def makecclib(atoms, popname="mulliken"):
    """Create cclib attributes and return a ccData from an ASE Atoms object.

    Available data (such as forces/gradients and potential energy/free
    energy) is assumed to be from SCF (see
    https://wiki.fysik.dtu.dk/ase/ase/atoms.html#adding-a-calculator).

    Bear in mind that ASE calculates temperature from the kinetic energy, so
    anything "static" (which includes anything cclib parses) will have zero
    temperature.

    Inputs:
        atoms - an instance of ASE `Atoms`
        popname - population analysis to use for atomic partial charges and
            atomic spin densities. Molecular charge and multiplicity are
            evaluated from them.
    """
    _check_ase(_found_ase)
    attributes = {
        "atomcoords": np.array([atoms.get_positions()]),
        "atomnos": atoms.get_atomic_numbers(),
        "atommasses": atoms.get_masses(),
        "natom": atoms.get_global_number_of_atoms(),
    }

    try:
        attributes["atomcharges"] = {popname: atoms.get_charges()}
    except (PropertyNotImplementedError, RuntimeError):
        attributes["atomcharges"] = {popname: atoms.get_initial_charges()}
    try:
        attributes["atomspins"] = {popname: atoms.get_magnetic_moments()}
    except (PropertyNotImplementedError, RuntimeError):
        attributes["atomspins"] = {
            popname: atoms.get_initial_magnetic_moments()
        }

    # the following is how ASE determines charge and multiplicity from initial
    # charges and initial magnetic moments in its Gaussian interface
    # (https://gitlab.com/ase/ase/-/blob/a26bda2160527ca7afc0135c69e4367a5bc5a264/ase/io/gaussian.py#L105)
    attributes["charge"] = attributes["atomcharges"][popname].sum()
    attributes["mult"] = attributes["atomspins"][popname].sum() + 1

    try:
        attributes["scfenergies"] = np.array([atoms.get_potential_energy()])
    except RuntimeError:
        pass
    try:
        attributes["grads"] = (
            -np.array([atoms.get_forces()]) * units.Bohr / units.Hartree
        )
    except RuntimeError:
        pass
    try:
        attributes["moments"] = [
            atoms.get_center_of_mass(),
            atoms.get_dipole_moment() / units.Bohr,
        ]
    except RuntimeError:
        pass
    try:
        attributes["freeenergy"] = (
            atoms.get_potential_energy(force_consistent=True) / units.Hartree
        )
    except RuntimeError:
        pass

    attributes["temperature"] = atoms.get_temperature()

    return ccData(attributes)


del find_package
